(*  Title:      Nitpick/Tools/nitpick.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Finite model generation for HOL formulas using Kodkod.
*)

signature NITPICK =
sig
  type hol_term = Term.term
  datatype mood = Optimistic | Realistic | Pessimistic

  type params = {
    cards_assigns: (typ option * int list) list,
    mults_assigns: ((string * typ) option * int list) list,
    iters_assigns: ((string * typ) option * int list) list,
    lockstep: bool,
    sat_solver: string,
    batch_size: int,
    blocking: bool,
    falsify: bool,
    normalize: bool,
    inductive_mood: mood,
    special_depth: int,
    skolem_depth: int,
    peephole_optim: bool,
    timeout: Time.time,
    auto_timeout: Time.time,
    tac_timeout: Time.time,
    sym_break: int,
    sharing_depth: int,
    flatten_props: bool,
    max_threads: int,
    verbose: bool,
    debug: bool,
    show_skolems: bool,
    show_datatypes: bool,
    show_consts: bool,
    eval: hol_term list,
    max_potential: int,
    check_potential: bool,
    check_genuine: bool,
    expect: string}

  val pick_nits_in_term :
    Proof.state -> params -> bool -> hol_term -> string * Proof.state
  val pick_nits_in_subgoal :
    Proof.state -> params -> bool -> int -> string * Proof.state
end;

structure Nitpick : NITPICK =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep
open NitpickTerm
open NitpickKodkod
open NitpickModel

type params = {
  cards_assigns: (typ option * int list) list,
  mults_assigns: ((string * typ) option * int list) list,
  iters_assigns: ((string * typ) option * int list) list,
  lockstep: bool,
  sat_solver: string,
  batch_size: int,
  blocking: bool,
  falsify: bool,
  normalize: bool,
  inductive_mood: mood,
  special_depth: int,
  skolem_depth: int,
  peephole_optim: bool,
  timeout: Time.time,
  auto_timeout: Time.time,
  tac_timeout: Time.time,
  sym_break: int,
  sharing_depth: int,
  flatten_props: bool,
  max_threads: int,
  verbose: bool,
  debug: bool,
  show_skolems: bool,
  show_datatypes: bool,
  show_consts: bool,
  eval: hol_term list,
  max_potential: int,
  check_potential: bool,
  check_genuine: bool,
  expect: string}

type problem_extension = {
  free_names: term list,
  ctor_names: term list,
  non_ctor_names: term list,
  rel_table: term NameTable.table,
  axioms: Kodkod.formula list,
  core: Kodkod.formula}

type rich_problem = Kodkod.problem * problem_extension

(* Proof.context -> string -> Term.term list -> Pretty.T *)
fun pretty_formulas _ _ [] = Pretty.str ""
  | pretty_formulas ctxt s ts =
    Pretty.chunks
        [Pretty.str (s ^ plural_s ts ^ ":"),
         Pretty.indent indent_size (Pretty.chunks
             (map2 (fn j => fn t =>
                       Pretty.block [Syntax.pretty_term ctxt t,
                                     Pretty.str (if j = 1 then "." else ";")])
                   (length ts downto 1) ts))]

(* Proof.state -> params -> bool -> hol_term -> string * Proof.state *)
fun pick_nits_in_term state params auto orig_t =
  let
    val thy = Proof.theory_of state
    val ctxt = Proof.context_of state

    val timer = Timer.startRealTimer ()
    val {cards_assigns, mults_assigns, iters_assigns, lockstep, sat_solver,
         batch_size, blocking, falsify, normalize, inductive_mood,
         special_depth, skolem_depth, peephole_optim, timeout, auto_timeout,
         tac_timeout, sym_break, sharing_depth, flatten_props, max_threads,
         verbose, debug, show_skolems, show_datatypes, show_consts, eval,
         max_potential, check_potential, check_genuine, expect} = params
    val (blocking, timeout, verbose, debug, max_potential) =
      if auto then (false, auto_timeout, false, false, 0)
      else (blocking, timeout, verbose, debug, max_potential)

    val state_ref = ref state
    (* Pretty.T -> unit *)
    val pprint =
      if auto then
        change state_ref o Proof.goal_message o K o Pretty.chunks
        o cons (Pretty.str "") o single o Pretty.mark Markup.hilite
      else
        priority o Pretty.string_of
    (* (() -> Pretty.T) -> unit *)
    fun pprint_m f = if auto then () else pprint (f ())
    fun pprint_v f = if verbose then pprint (f ()) else ()
    fun pprint_d f = if debug then pprint (f ()) else ()
    (* string -> unit *)
    val print = pprint o Pretty.block o pstrs
    (* (() -> string) -> unit *)
    fun print_m f = pprint_m (Pretty.block o pstrs o f)
    fun print_v f = pprint_v (Pretty.block o pstrs o f)
    fun print_d f = pprint_d (Pretty.block o pstrs o f)

    val _ = print_m (K "Nitpicking...")
    val original_max_potential = max_potential
(*
    val _ = tracing ("*** " ^ Syntax.string_of_term ctxt orig_t)
*)
    val (defs, nondefs) = all_definitional_and_nondefinitional_axioms_of thy
    val def_table = ref (const_def_table ctxt defs)
    val nondef_table = ref (const_nondef_table nondefs)
    val simp_table = ref (const_simp_table ctxt)
    val psimp_table = ref (const_psimp_table ctxt)
    val intro_table = ref (inductive_intro_table ctxt)
    val (ext_ctxt as {skolems, special_funs, wfs, ...}) =
      {thy = thy, ctxt = ctxt, inductive_mood = inductive_mood,
       special_depth = special_depth, skolem_depth = skolem_depth,
       tac_timeout = tac_timeout, eval = eval, def_table = def_table,
       nondef_table = nondef_table, simp_table = simp_table,
       psimp_table = psimp_table, intro_table = intro_table, skolems = ref [],
       special_funs = ref [], better_founded_preds = ref [], wfs = ref []}
    val _ = null (Term.add_tvars orig_t [])
            orelse raise NotSupported ("pick_nits_in_term",
                                       "schematic type variables")
    val vars = sort_wrt (fst o fst) (Term.add_vars orig_t [])
    val norm_t = the (try (normalize ? Nbe.norm_term thy) orig_t)
                 handle Option => orig_t
    val _ =
      if norm_t <> orig_t then
        pprint_m (fn () => pretty_formulas ctxt "Normalized formula" [norm_t])
      else
        ()
    val ex_closed_t =
      fold (fn ((s, j), T) => fn t =>
               HOLogic.exists_const T
               $ Abs (s, T, abstract_over (Var ((s, j), T), t)))
           vars norm_t
    val (axiom_ts, core_t) =
      (the (try (normalize ? Nbe.norm_term thy) ex_closed_t)
       handle Option => ex_closed_t)
      |> preprocess_hol_term ext_ctxt falsify

    (* (string * typ) * bool -> unit *)
    fun print_wf (x, wf) =
      pprint (Pretty.blk (0,
          pstrs "The inductively defined predicate \"" @
          Syntax.pretty_term ctxt (Const x) ::
          pstrs (if wf then
                   "\" was proved wellfounded. Nitpick can compute it \
                   \efficiently."
                 else
                   " could not be proved wellfounded. Nitpick will unroll \
                   \it.")))
    val _ = if verbose then List.app print_wf (!wfs) else ()
    val _ =
      pprint_d (fn () => Pretty.chunks
                          [pretty_formulas ctxt "Preprocessed formula" [core_t],
                           pretty_formulas ctxt "Relevant axiom" axiom_ts])
    val Ts = ground_types_in_hol_term ctxt (core_t :: axiom_ts)
(*
    val _ = tracing "Types:"
    val _ = map (tracing o Syntax.string_of_typ ctxt) Ts
    val _ = tracing "Axioms:"
    val _ = map (tracing o Syntax.string_of_term ctxt) axiom_ts
    val _ = tracing "Core:"
    val _ = tracing (Syntax.string_of_term ctxt core_t)
*)
    val axiom_us = map (from_hol_term thy StrictEq) axiom_ts
    val core_u = from_hol_term thy Eq core_t
    val free_names = fold add_free_names (core_u :: axiom_us) []
    val const_names = fold add_const_names (core_u :: axiom_us) []
    val non_ctor_names = filter (not o is_ctor_name thy) const_names
    val dodgy_ind_pred_names =
      if inductive_mood = Optimistic then
        filter (fn ConstName (s, T, _) => is_inductive_pred ext_ctxt (s, T)
                 | _ => false) non_ctor_names
      else
        []
(*
    val _ = map (tracing o string_for_term ctxt) axiom_us
    val _ = tracing (string_for_term ctxt core_u)
*)
    val _ = if sat_solver = "smart" then
              print_v (fn () =>
                          "Using SAT solver " ^
                          quote (KodkodSAT.smart_sat_solver_name ()) ^
                          ". The following solvers are configured: " ^
                          commas (map quote
                                   (KodkodSAT.configured_sat_solvers ())) ^ ".")
            else
              ()
    (* bool -> scope -> rich_problem option *)
    fun problem_for_scope max_potential scope =
      let
(*
        val _ = tracing (string_for_scope "; " scope)
*)
        val dtypes = #datatypes scope
        val all_precise = forall (is_precise_type dtypes) Ts
        val nat_card = card_of_type (#card_assigns scope) nat_T
        val int_card = card_of_type (#card_assigns scope) int_T
        val kk = kodkod_ctors peephole_optim nat_card int_card
        val (free_names, rep_table) =
          choose_reps_for_free_vars scope free_names NameTable.empty
        val (ctor_names, rep_table) = choose_reps_for_all_ctors scope rep_table
        val (non_ctor_names, rep_table) =
          choose_reps_for_consts scope all_precise non_ctor_names rep_table
        val axiom_us =
          map (choose_reps_in_term true Formula scope rep_table) axiom_us
        val core_u = choose_reps_in_term false Formula scope rep_table core_u
(*
        val _ = map (fn x => tracing (string_for_term ctxt x))
                    (free_names @ const_names)
        val _ = map (tracing o string_for_term ctxt) axiom_us
        val _ = tracing (string_for_term ctxt core_u)
*)
        val (free_rels, pool, rel_table) =
          rename_free_vars free_names initial_pool NameTable.empty
        val (ctor_rels, pool, rel_table) =
          rename_free_vars ctor_names pool rel_table
        val (non_ctor_rels, pool, rel_table) =
          rename_free_vars non_ctor_names pool rel_table
        val core_u = rename_vars_in_term pool rel_table core_u
        val axiom_us = map (rename_vars_in_term pool rel_table) axiom_us
(*
        val _ = map (fn x => tracing (string_for_term ctxt x))
                    (free_rels @ ctor_rels @ non_ctor_rels)
        val _ = map (tracing o string_for_term ctxt) axiom_us
        val _ = tracing (string_for_term ctxt core_u)
*)
        val axiom_fs =
          map (to_kodkod_formula dtypes peephole_optim false true false kk)
              axiom_us
        val core_f = to_kodkod_formula dtypes peephole_optim (max_potential > 0)
                                       (max_potential > 0) falsify kk core_u
        val formula = fold s_and axiom_fs core_f
        val comment = PrintMode.setmp [] multiline_string_for_scope scope
        val kodkod_sat_solver = KodkodSAT.sat_solver_spec sat_solver |> snd
        val settings = [("solver", commas (map quote kodkod_sat_solver)),
                        ("skolem_depth", "-1"),
                        ("bit_width", "16"),
                        ("symmetry_breaking", Int.toString sym_break),
                        ("sharing", Int.toString sharing_depth),
                        ("flatten", Bool.toString flatten_props)]
        val plain_rels = free_rels @ non_ctor_rels
        val plain_bounds = map bound_for_plain_rel plain_rels
        val plain_axioms = map (axiom_for_plain_rel peephole_optim kk)
                               plain_rels
        val ctor_bounds = map (bound_for_ctor_rel dtypes) ctor_rels
        val ctor_axioms = map (axiom_for_ctor_rel peephole_optim dtypes kk)
                              ctor_rels
        val nfa_table = map (nfa_entry_for_dtype kk dtypes rel_table) dtypes
        val dtype_axioms = map (acyclicity_axiom kk nfa_table o #typ) dtypes
        val low_level_axioms = plain_axioms @ ctor_axioms @ dtype_axioms
        val univ_card =
          univ_card nat_card int_card (plain_bounds @ ctor_bounds) formula
        val built_in_bounds =
          bounds_for_built_in_rels_in_formula (max_potential > 0) univ_card
                                              nat_card int_card formula
        val bounds =
          merge_bounds (built_in_bounds @ plain_bounds @ ctor_bounds) []
        val formula = fold_rev s_and low_level_axioms formula
      in
        SOME ({comment = comment, settings = settings, univ_card = univ_card,
               tuple_assigns = [], bounds = bounds, int_bounds = [],
               expr_assigns = [], formula = formula},
              {free_names = free_names, ctor_names = ctor_names,
               non_ctor_names = non_ctor_names, rel_table = rel_table,
               axioms = axiom_fs @ low_level_axioms, core = core_f})
      end
      handle Limit _ => NONE

    (* int -> (''a * int list list) list -> ''a -> Kodkod.tuple_set *)
    fun tuple_set_for_rel univ_card =
      Kodkod.TupleSet o map (kk_tuple univ_card) o the oo AList.lookup op=
    (* int -> term list -> Kodkod.raw_bound list -> Kodkod.tuple_set list
       -> Kodkod.n_ary_index * string
       -> (Kodkod.n_ary_index * string) list * Kodkod.tuple_set list *)
    fun debase_simple_bound univ_card [] raw_bounds _ (x, s) =
        ([(x, s)], [tuple_set_for_rel univ_card raw_bounds x])
      | debase_simple_bound univ_card (r :: rs) raw_bounds tss (x, s) =
        (case r of
           FreeRel (x', _, R, _) =>
           if x = x' then
             ([(x, s)],
              if body_rep R = Formula then
                [Kodkod.TupleSet [], tuple_set_for_rel univ_card raw_bounds x]
              else
                tss)
           else
             debase_simple_bound univ_card rs raw_bounds tss (x, s)
         | _ => raise BadTerm ("debase_simple_bound", [r]))
    (* int -> term list -> Kodkod.raw_bound list
       -> (Kodkod.n_ary_index * string) list * Kodkod.tuple_set list
       -> ((Kodkod.n_ary_index * string) list * Kodkod.tuple_set list) list *)
    fun debase_bound univ_card dodgy_rels raw_bounds (zs, tss) =
      map (debase_simple_bound univ_card dodgy_rels raw_bounds tss) zs

    (* rich_problem -> Kodkod.raw_bound list -> Kodkod.problem *)
    fun debase_inductive_problem ({comment, settings, univ_card, tuple_assigns,
                                   bounds, int_bounds, expr_assigns, ...},
                                  {rel_table, axioms, core, ...}) raw_bounds =
      let val dodgy_rels = map (lookup_name rel_table) dodgy_ind_pred_names in
        {comment = "debased\n" ^ comment, settings = settings,
         univ_card = univ_card, tuple_assigns = tuple_assigns,
         bounds = List.concat (map (debase_bound univ_card dodgy_rels
                                                 raw_bounds) bounds),
         int_bounds = int_bounds, expr_assigns = expr_assigns,
         formula = fold s_and axioms (Kodkod.Not core)}
      end

    val the_word_model = if falsify then "counterexample" else "model"

    datatype model_kind = Genuine | Potential | GenuineModIndPreds

    val met_potential = ref false

    (* scope -> problem_extension -> model_kind -> raw_bound list
       -> bool option *)
    fun print_and_check_model scope
            {free_names, ctor_names, non_ctor_names, rel_table, ...}
            kind bounds =
      (pprint (Pretty.chunks
           [Pretty.blk (0,
                (pstrs ("Nitpick found a " ^
                        (if kind = Genuine then "genuine" else "potential") ^
                        " " ^ the_word_model) @
                 (case pretties_for_scope scope verbose of
                    [] => []
                  | pretties => pstrs " for " @ pretties) @
                 [Pretty.str ":\n"])),
            Pretty.indent indent_size
                (reconstruct_hol_model
                     {show_skolems = show_skolems,
                      show_datatypes = show_datatypes,
                      show_consts = show_consts}
                     scope eval (!skolems) (!special_funs) free_names ctor_names
                     non_ctor_names rel_table bounds)]);
       if kind = Genuine then
         if check_genuine then
           ((case prove_hol_model scope tac_timeout free_names ctor_names
                                  rel_table bounds falsify orig_t of
              SOME true => print ("Confirmation by \"auto\": The above " ^
                                  the_word_model ^ " is really genuine.")
            | SOME false =>
              error ("A supposedly genuine " ^ the_word_model ^ " was shown to \
                     \be spurious by \"auto\".\nThis should never happen.\n\
                     \Please send a bug report to blanchette\
                     \@in.tum.de.")
            | NONE => print "No confirmation by \"auto\"."); NONE)
         else
           NONE
       else
         if kind = Potential then
           (met_potential := true;
            if check_potential then
              let
                val status = prove_hol_model scope tac_timeout free_names
                                             ctor_names rel_table bounds falsify
                                             orig_t
              in
                (case status of
                   SOME true => print ("Confirmation by \"auto\": The above " ^
                                       the_word_model ^ " is genuine.")
                 | SOME false => print ("Refutation by \"auto\": The above " ^
                                        the_word_model ^ " is spurious.")
                 | NONE => print "No confirmation by \"auto\".");
                status
              end
            else
              NONE)
         else
           NONE)
    (* int * bool -> scope list -> Kodkod.problem list
       -> (int * bool) option *)
    fun solve_any_problem (max_potential, donno) scopes problems =
      let
       (* int -> model_kind -> raw_bound list -> bool option *)
        fun print_and_check j =
          print_and_check_model (nth scopes j) (snd (nth problems j))
        (* int -> raw_bound list -> raw_bound list -> term -> bool *)
        fun is_problematic_inductive_pred j raw_bounds1 raw_bounds2 name =
          let val {rel_table, ...} = snd (nth problems j) in
            tuple_list_for_name rel_table raw_bounds1 name
            <> tuple_list_for_name rel_table raw_bounds2 name
          end
        (* term list -> unit *)
        fun handle_problematic_inductive_preds names =
          let
            val ss = sort_distinct string_ord
                         (map (quote o original_name o nickname_of) names)
          in
            print_m (fn () => "The inductive predicate" ^ plural_s ss ^ " " ^
                              space_implode " " (serial_commas ss) ^
                              " might be non-wellfounded. Bailing out. Try \
                              \again with \"inductive_mood\" set to \
                              \\"realistic\" or \"pessimistic\"."); NONE
          end
        (* int -> raw_bound list -> (int * bool) option *)
        fun handle_genuine_model_mod_inductive_preds j bounds =
          if null dodgy_ind_pred_names then
            (print_and_check j Genuine bounds; NONE)
          else
            (if max_potential > 0 then
               print_and_check j GenuineModIndPreds bounds
             else
               NONE;
             case Kodkod.solve_any_problem timeout max_threads
                      [debase_inductive_problem (nth problems j) bounds] of
               Kodkod.Satisfiable (_, new_bounds) =>
               handle_problematic_inductive_preds
                   (filter (is_problematic_inductive_pred j bounds new_bounds)
                           dodgy_ind_pred_names)
             | Kodkod.Unsatisfiable =>
               if max_potential > 0 then
                 (print ("Confirmation by Kodkod: The above " ^ the_word_model ^
                         " is genuine."); NONE)
               else
                 (print_and_check j Genuine bounds; NONE)
             | _ => handle_problematic_inductive_preds dodgy_ind_pred_names)
      in
        case Kodkod.solve_any_problem timeout max_threads (map fst problems) of
          Kodkod.Satisfiable (j, bounds) =>
          if is_only_liberally_satisfiable bounds then
            if print_and_check j Potential bounds = SOME true then
              NONE
            else
              solve_any_problem (max_potential - 1, donno)
                  (List.drop (scopes, j))
                  (if max_potential > 1 then
                     apfst illiberalize_problem (nth problems j) ::
                     List.drop (problems, j + 1)
                   else
                     map (apfst illiberalize_problem) (List.drop (problems, j)))
          else
            handle_genuine_model_mod_inductive_preds j bounds
        | Kodkod.Unsatisfiable => SOME (max_potential, donno)
        | Kodkod.Error s =>
          (if verbose orelse not donno then warning ("Kodkod error: " ^ s ^ ".")
           else ();
           SOME (max_potential, true))
      end

    (* int * bool -> scope list -> (int * bool) option *)
    fun run_batch (max_potential, donno) scopes =
      let
        val _ =
          if null scopes then
            print_m (K "The scope specification is inconsistent.")
          else if verbose then
            pprint (Pretty.chunks
                [Pretty.blk (0,
                     pstrs ("Trying " ^ Int.toString (length scopes) ^
                            " scope" ^ plural_s scopes ^ ":")),
                 Pretty.indent indent_size
                     (Pretty.chunks (map2
                          (fn j => fn scope =>
                              Pretty.block (pretties_for_scope scope true @
                                  [Pretty.str (if j = 1 then "." else ";")]))
                          (length scopes downto 1) scopes))])
          else
            ()
        val problems = List.mapPartial (problem_for_scope max_potential) scopes
      in
        solve_any_problem (max_potential,
                           donno orelse length problems < length scopes)
                          scopes problems
      end

    (* int * bool -> scope list -> Kodkod.outcome *)
    fun run_batches (max_potential, donno) [] =
        if donno then
          (print_m (K "Nitpick encountered some problem. The search is \
                      \inconclusive."); "unknown")
        else if max_potential = original_max_potential then
          (print_m (K ("Nitpick found no " ^ the_word_model ^ ".")); "none")
        else
          (print_m (K ("Nitpick could not find a better " ^ the_word_model ^
                       ".")); "potential")
      | run_batches (max_potential, donno) (batch :: batches) =
        (case run_batch (max_potential, donno) batch of
           SOME (max_potential, donno) =>
           run_batches (max_potential, donno) batches
         | NONE => "genuine")

    (* unit -> string *)
    fun run_the_batches () =
      (met_potential := false;
       run_batches (max_potential, false)
                   (batch_list batch_size
                        (all_scopes ctxt cards_assigns mults_assigns
                                    iters_assigns Ts lockstep)))

    val outcome_code =
      time_limit timeout run_the_batches ()
      handle TimeLimit.TimeOut =>
             (print_m (K "Nitpick ran out of time.");
              if !met_potential then "potential" else "unknown")
    val _ =
      print_v (fn () => "Total time: " ^
                        Int.toString (Time.toMilliseconds
                              (Timer.checkRealTimer timer)) ^ " ms.")
  in
    if expect = "" orelse outcome_code = expect then (outcome_code, !state_ref)
    else error ("Unexpected outcome: " ^ quote outcome_code ^ ".")
  end

(* Proof.state -> params -> thm -> int -> string * Proof.state *)
fun pick_nits_in_subgoal state params auto subgoal =
  let
    val ctxt = Proof.context_of state
    val thm = snd (snd (Proof.get_goal state))
    val assms = map term_of (Assumption.assms_of ctxt)
    val (t, frees) = Logic.goal_params (prop_of thm) subgoal
    val t = Logic.list_implies (assms, subst_bounds (frees, t))
  in pick_nits_in_term state params auto t end

end;
