(*  Title:      Nitpick/Tools/nitpick_hol.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Auxiliary HOL-related functions used by Nitpick.
*)

signature NITPICK_HOL =
sig
  type hol_term = Term.term
  type const_table = hol_term list Symtab.table
  type special_fun =
    ((string * typ) * int list * hol_term list) * (string * typ)
  type better_founded = (string * typ) * (string * typ)
  type wf_cache = ((string * typ) * bool) list

  datatype mood = Optimistic | Realistic | Pessimistic

  type extended_context = {
    thy: theory,
    ctxt: Proof.context,
    inductive_mood: mood,
    special_depth: int,
    skolem_depth: int,
    tac_timeout: Time.time,
    eval: hol_term list,
    def_table: const_table ref,
    nondef_table: const_table ref,
    simp_table: const_table ref,
    psimp_table: const_table ref,
    intro_table: const_table ref,
    skolems: (string * string list) list ref,
    special_funs: special_fun list ref,
    better_founded_preds: better_founded list ref,
    wfs: wf_cache ref}

  val numeral_prefix : string
  val skolem_prefix : string
  val eval_prefix : string
  val is_TFree : typ -> bool
  val is_higher_order_type : typ -> bool
  val is_fun_type : typ -> bool
  val is_set_type : typ -> bool
  val is_pair_type : typ -> bool
  val is_iterator_type : typ -> bool
  val is_integer_type : typ -> bool
  val iterator_type_for_const : string * typ -> typ
  val const_for_iterator_type : typ -> string * typ
  val second_range_type : typ -> typ
  val third_range_type : typ -> typ
  val is_typedef : theory -> string -> bool
  val is_datatype_like : theory -> typ -> bool
  val is_ctor : theory -> string * typ -> bool
  val original_name : string -> string
  val user_friendly_const :
    hol_term list -> (string * string list) list -> special_fun list
    -> string * typ -> hol_term
  val instantiate_type : theory -> typ -> typ -> typ -> typ
  val eta_expand : typ list -> hol_term -> int -> hol_term
  val extensionalize : hol_term -> hol_term
  val all_definitional_and_nondefinitional_axioms_of :
    theory -> hol_term list * hol_term list
  val const_def_table : Proof.context -> hol_term list -> const_table
  val const_nondef_table : hol_term list -> const_table
  val const_simp_table : Proof.context -> const_table
  val const_psimp_table : Proof.context -> const_table
  val inductive_intro_table : Proof.context -> const_table
  val is_inductive_pred : extended_context -> string * typ -> bool
  val preprocess_hol_term :
    extended_context -> bool -> hol_term -> hol_term list * hol_term
  val ground_types_in_hol_term : Proof.context -> hol_term list -> typ list
end;

structure NitpickHOL : NITPICK_HOL =
struct

open NitpickUtil

type const_table = hol_term list Symtab.table
type special_fun = ((string * typ) * int list * hol_term list) * (string * typ)
type better_founded = (string * typ) * (string * typ)
type wf_cache = ((string * typ) * bool) list

datatype mood = Optimistic | Realistic | Pessimistic

type extended_context = {
  thy: theory,
  ctxt: Proof.context,
  inductive_mood: mood,
  special_depth: int,
  skolem_depth: int,
  tac_timeout: Time.time,
  eval: hol_term list,
  def_table: const_table ref,
  nondef_table: const_table ref,
  simp_table: const_table ref,
  psimp_table: const_table ref,
  intro_table: const_table ref,
  skolems: (string * string list) list ref,
  special_funs: special_fun list ref,
  better_founded_preds: better_founded list ref,
  wfs: wf_cache ref}

val numeral_prefix = nitpick_prefix ^ "num_"
val iterator_prefix = nitpick_prefix ^ "it_"
val wf_prefix = nitpick_prefix ^ "wf_"
val skolem_prefix = nitpick_prefix ^ "sk"
val special_prefix = nitpick_prefix ^ "sp"
val eval_prefix = nitpick_prefix ^ "eval"

(* int -> string *)
fun skolem_prefix_for j = skolem_prefix ^ nat_subscript j ^ "_"
fun special_prefix_for j = special_prefix ^ nat_subscript j ^ "_"

val built_in_consts =
  [(@{const_name all}, 1),
   (@{const_name "=="}, 2),
   (@{const_name "==>"}, 2),
   (@{const_name Trueprop}, 1),
   (@{const_name Not}, 1),
   (@{const_name False}, 0),
   (@{const_name True}, 0),
   (@{const_name All}, 1),
   (@{const_name Ex}, 1),
   (@{const_name "op ="}, 2),
   (@{const_name "op &"}, 2),
   (@{const_name "op |"}, 2),
   (@{const_name "op -->"}, 2),
   (@{const_name If}, 3),
   (@{const_name Let}, 2),
   (@{const_name Unity}, 0),
   (@{const_name Pair}, 2),
   (@{const_name fst}, 1),
   (@{const_name snd}, 1),
   (@{const_name Id}, 0),
   (@{const_name Set.empty}, 0),
   (@{const_name Set.UNIV}, 0),
   (@{const_name "op :"}, 2),
   (@{const_name insert}, 2),
   (@{const_name trancl}, 1),
   (@{const_name rtrancl}, 1),
   (@{const_name Set.Un}, 2),
   (@{const_name Set.Int}, 2),
   (@{const_name Suc}, 0),
   (@{const_name finite}, 1),
   (@{const_name card}, 1),
   (@{const_name nat}, 0)]
val built_in_typed_consts =
  [((@{const_name uminus_class.uminus}, int_T --> int_T), 0),
   ((@{const_name of_nat}, nat_T --> int_T), 0)]
val built_in_set_consts =
  [(@{const_name minus_class.minus}, 2),
   (@{const_name ord_class.less}, 2),
   (@{const_name ord_class.less_eq}, 2),
   (@{const_name uminus_class.uminus}, 2)]
val built_in_integer_consts =
  [(@{const_name zero_class.zero}, 0),
   (@{const_name one_class.one}, 0),
   (@{const_name plus_class.plus}, 0),
   (@{const_name minus_class.minus}, 0),
   (@{const_name times_class.times}, 0),
   (@{const_name ord_class.less}, 2),
   (@{const_name ord_class.less_eq}, 2)]

(* typ -> bool *)
fun is_TFree (TFree _) = true
  | is_TFree _ = false
fun is_higher_order_type (Type ("fun", _)) = true
  | is_higher_order_type (Type (_, Ts)) = exists is_higher_order_type Ts
  | is_higher_order_type _ = false
fun is_fun_type (Type ("fun", _)) = true
  | is_fun_type _ = false
fun is_set_type (Type ("fun", [_, Type ("bool", [])])) = true
  | is_set_type _ = false
fun is_pair_type (Type ("*", _)) = true
  | is_pair_type _ = false
fun is_iterator_type (Type (s, _)) = String.isPrefix iterator_prefix s
  | is_iterator_type _ = false
fun is_integer_type T = (T = nat_T orelse T = int_T orelse is_iterator_type T)

(* string * typ -> typ *)
fun iterator_type_for_const (s, T) = Type (iterator_prefix ^ s, [domain_type T])
(* typ -> string * typ *)
fun const_for_iterator_type (Type (s, [T])) =
    (unprefix iterator_prefix s, T --> bool_T)
  | const_for_iterator_type T = raise BadType ("const_for_iterator_type", [T])

(* typ -> typ *)
val second_range_type = range_type o range_type
val third_range_type = range_type o second_range_type

(* theory -> string -> bool *)
val is_typedef = isSome oo TypedefPackage.get_info
val is_datatype = isSome oo DatatypePackage.get_datatype
(* theory -> typ -> bool *)
fun is_pure_typedef thy (T as Type (s, _)) =
    is_typedef thy s andalso not (is_datatype thy s)
    andalso not (is_integer_type T)
  | is_pure_typedef _ _ = false
fun is_datatype_like thy (T as Type (s, _)) =
    (is_typedef thy s orelse T = @{typ ind}) andalso not (is_integer_type T)
  | is_datatype_like _ _ = false

(* theory -> string * typ -> bool * bool *)
fun is_abs_and_rep_fun thy (s, Type ("fun", [_, Type (s', _)])) =
    (case TypedefPackage.get_info thy s' of
       SOME {Abs_name, Rep_name, ...} => (s = Abs_name, s = Rep_name)
     | NONE => (false, false))
  | is_abs_and_rep_fun _ _ = (false, false)

(* theory -> string * typ -> bool *)
fun is_ctor thy (x as (s, T)) =
  Refute.is_IDT_constructor thy x
  orelse (fst (is_abs_and_rep_fun thy x)
          andalso is_pure_typedef thy (range_type T))
  orelse s = @{const_name Zero_Rep} orelse s = @{const_name Suc_Rep}

(* string -> string *)
fun original_name s =
  if String.isPrefix nitpick_prefix s then
    original_name (Substring.string (Substring.triml 1
        (snd (Substring.position "_" (Substring.full s)))))
  else
    s

(* (string * int) * typ -> hol_term -> hol_term *)
fun abs_var ((s, j), T) body = Abs (s, T, abstract_over (Var ((s, j), T), body))

(* hol_term list -> (string * string list) list -> special_fun list
   -> string * typ -> hol_term *)
fun user_friendly_const eval skolems special_funs (s, T) =
  if String.isPrefix special_prefix s then
    let
      val ((s', T'), js, ts) = the_single (AList.find op= special_funs (s, T))
      val max_j = List.last js
      val Ts = List.take (binder_types T', max_j + 1)
      val vars = fold Term.add_vars ts []
                 |> sort (fast_string_ord o pairself (fst o fst)) |> rev
      val ts' = map2 (fn T => AList.lookup op= (js ~~ ts)
                              #> the_default (Const ("\\<star>", T)))
                     Ts (0 upto max_j)
    in
      list_comb (user_friendly_const eval skolems special_funs (s', T'), ts')
      |> fold abs_var vars
    end
  else if String.isPrefix wf_prefix s then
    lambda (Free ("i", nat_T)) (Const (original_name s, range_type T))
  else if String.isPrefix skolem_prefix s then
    let
      val (binder_Ts, body_T) = strip_type T
      val ss = the (AList.lookup op= skolems s)
      val (Ts, Ts') = chop (length ss) binder_Ts
      val frees = map Free (ss ~~ Ts)
    in fold lambda frees (Const (original_name s, Ts' ---> T)) end
  else if String.isPrefix eval_prefix s then
    nth eval (the (Int.fromString (unprefix eval_prefix s)))
  else
    Const (original_name s, T)

(* typ -> typ *)
val varify_type = map_type_tfree (fn (s, S) => TVar ((s, 0), S))

(* theory -> typ -> typ -> typ -> typ *)
fun instantiate_type thy T1 T1' T2 =
  Envir.typ_subst_TVars (Sign.typ_match thy (varify_type T1, T1') Vartab.empty)
                        (varify_type T2)
  handle Type.TYPE_MATCH => raise BadType ("instantiate_type", [T1, T1'])

(* typ list -> hol_term -> int -> hol_term *)
fun eta_expand _ t 0 = t
  | eta_expand Ts t n =
    fold_rev (curry3 Abs ("x\\<^isub>\\<eta>" ^ nat_subscript n))
             (List.take (binder_types (fastype_of1 (Ts, t)), n))
             (list_comb (incr_boundvars n t, map Bound (n - 1 downto 0)))

(* hol_term -> hol_term *)
fun extensionalize t =
  case t of
    @{const Trueprop} $ t1 => extensionalize t1
  | Const (@{const_name "op ="}, _) $ t1 $ Abs (s, T, t2) =>
    let val v = Var ((s, maxidx_of_term t + 1), T) in
      extensionalize (HOLogic.mk_eq (t1 $ v, subst_bound (v, t2)))
    end
  | _ => t

(* hol_term -> bool *)
fun is_arity_type_axiom
        (Const (@{const_name HOL.type_class}, _)
         $ Const (@{const_name TYPE}, _)) = true
  | is_arity_type_axiom _ = false
(* theory -> hol_term -> bool *)
fun is_typedef_axiom thy
        (@{const "==>"} $ _ $ (@{const Trueprop}
         $ (Const (@{const_name Typedef.type_definition}, _)
         $ Const (_, Type ("fun", [Type (s, _), _])) $ Const _ $ _))) =
    is_typedef thy s
  | is_typedef_axiom _ _ = false

(* Distinguishes between (1) constant definition axioms, (2) type arity and
   typedef axioms, and (3) other axioms, and returns the pair ((1), (3)).
   Type axioms are uninteresting to Nitpick, because it can retrieve them using
   "TypedefPackage.get_info". *)

(* theory -> (string * hol_term) list -> string list
   -> hol_term list * hol_term list *)
fun partition_axioms_by_definitionality thy axioms def_names =
  let
    val axioms = sort (fast_string_ord o pairself fst) axioms
    val defs =
      OrdList.inter (fast_string_ord o apsnd fst) def_names axioms
    val nondefs =
      OrdList.subtract (fast_string_ord o apsnd fst) def_names axioms
      |> filter_out ((is_arity_type_axiom orf is_typedef_axiom thy) o snd)
  in pairself (map snd) (defs, nondefs) end

(* theory -> hol_term list * hol_term list *)
fun all_definitional_and_nondefinitional_axioms_of thy =
  let
    val def_names =
      Defs.all_specifications_of (Theory.defs_of thy) |> maps snd
      |> filter #is_def |> map #name |> OrdList.make fast_string_ord
    val thys = thy :: Theory.ancestors_of thy
    val (main_thys, user_thys) =
      List.partition (fn thy => Theory.subthy (thy, @{theory Main})) thys
    val main_axioms = maps Theory.axioms_of main_thys
    val user_axioms = maps Theory.axioms_of user_thys
    val (main_defs, _) =
      partition_axioms_by_definitionality thy main_axioms def_names
    val (user_defs, user_nondefs) =
      partition_axioms_by_definitionality thy user_axioms def_names
  in (user_defs @ main_defs, user_nondefs) end

(* string * typ -> int option *)
fun arity_of_built_in_const (s, T) =
  case AList.lookup op= built_in_consts s of
    SOME n => SOME n
  | NONE =>
    case AList.lookup op= built_in_typed_consts (s, T) of
      SOME n => SOME n
    | NONE =>
      if is_fun_type T andalso is_set_type (domain_type T) then
        AList.lookup op= built_in_set_consts s
      else if is_integer_type T
              orelse (is_fun_type T
                      andalso is_integer_type (domain_type T)) then
        AList.lookup op= built_in_integer_consts s
      else
        NONE
(* string * typ -> bool *)
val is_built_in_const = isSome o arity_of_built_in_const

(* This function is designed to work for both real definition axioms and
   simplification rules (equational specifications). *)

(* hol_term -> (string * typ) option *)
fun const_under_def t =
  case t of
    @{const "==>"} $ _ $ t2 => const_under_def t2
  | Const (@{const_name "=="}, _) $ t1 $ _ => const_under_def t1
  | @{const Trueprop} $ t1 => const_under_def t1
  | Const (@{const_name "op ="}, _) $ t1 $ _ => const_under_def t1
  | _ => try (dest_Const o head_of) t

(* hol_term -> (string * hol_term) option *)
fun pair_for_prop t =
  case const_under_def t of
    SOME x => if is_built_in_const x then NONE else SOME (fst x, t)
  | _ => raise BadHOLTerm ("pair_for_prop", [t])

(* We add the "Nitpick" theorems to the tables if the "Nitpick" theory wasn't
   imported by the user. *)

(* (Proof.context -> hol_term list) -> Proof.context -> const_table *)
fun table_for get ctxt =
  (not (member Theory.eq_thy (Theory.ancestors_of (ProofContext.theory_of ctxt))
               (ProofContext.theory_of @{context}))
   ? append (get @{context})) (get ctxt)
  |> List.mapPartial pair_for_prop |> AList.group op= |> Symtab.make

(* Proof.context -> hol_term list -> const_table *)
fun const_def_table ctxt ts =
  table_for (map Thm.prop_of o Nitpick_Const_Def_Thms.get) ctxt
  |> fold (fn (s, t) => Symtab.map_default (s, []) (cons t))
          (List.mapPartial pair_for_prop ts)
(* hol_term list -> const_table *)
fun const_nondef_table ts =
  fold (fn t => append (map (fn s => (s, t)) (Term.add_const_names t []))) ts []
  |> AList.group op= |> Symtab.make
(* Proof.context -> const_table *)
val const_simp_table = table_for (map Thm.prop_of o Nitpick_Const_Simp_Thms.get)
val const_psimp_table =
  table_for (map Thm.prop_of o Nitpick_Const_Psimp_Thms.get)
val inductive_intro_table =
  table_for (map Thm.prop_of o Nitpick_Ind_Intro_Thms.get)

(* Here we crucially rely on "Refute.specialize_type" performing a preorder
   traversal of the term, without which the wrong occurrence of a constant could
   be matched in the face of overloading. *)

(* theory -> const_table -> string * typ -> hol_term list *)
fun def_props_for_const thy table (x as (s, _)) =
  these (Symtab.lookup table s)
  |> List.mapPartial (try (Refute.specialize_type thy x))
  |> filter (equal (SOME x) o const_under_def)

(* Similar to "Refute.specialize_type" but returns all matches rather than only
   the first (preorder) match. *)

(* theory -> string * typ -> hol_term -> hol_term list *)
fun multi_specialize_type thy (x as (s, T)) t =
  let
    (* hol_term -> (typ * hol_term) list -> (typ * hol_term) list *)
    fun aux (Const (s', T')) ys =
        if s = s' then
          (if AList.defined op= ys T' then
            I
          else
            cons (T', Refute.monomorphic_term
                          (Sign.typ_match thy (T', T) Vartab.empty) t)
            handle Type.TYPE_MATCH => I) ys
        else
          ys
      | aux _ ys = ys
  in map snd (fold_aterms aux t []) end

(* theory -> const_table -> string * typ -> hol_term list *)
fun nondef_props_for_const thy table (x as (s, _)) =
  these (Symtab.lookup table s) |> maps (multi_specialize_type thy x)
  handle Refute.REFUTE _ =>
  raise NotSupported ("nondef_props_for_const",
                      "too much polymorphism in axiom involving " ^ quote s)


(* string -> string -> string *)
fun prefix_with_const_name s = prefix (List.last (space_explode "." s) ^ ".")
(* string -> string *)
fun unprefix_with_const_name s =
  List.last (space_explode "." s) handle Empty => ""
(* string -> hol_term -> hol_term *)
val prefix_abs_vars_with_const_name = Term.map_abs_vars o prefix_with_const_name
(* hol_term -> hol_term *)
val unprefix_abs_vars_with_const_names =
  Term.map_abs_vars unprefix_with_const_name

(* extended_context -> string * typ -> hol_term option *)
fun def_of_const (ext_ctxt as {thy, def_table, ...}) (x as (s, _)) =
  (if String.isPrefix nitpick_prefix s then NONE
   else try (Refute.norm_rhs o hd) (def_props_for_const thy (!def_table) x))
  |> Option.map (prefix_abs_vars_with_const_name s)

datatype fixpoint_kind = Lfp | Gfp | NoFp

(* hol_term -> fixpoint_kind *)
fun fixpoint_kind_of_rhs (Abs (_, _, t)) = fixpoint_kind_of_rhs t
  | fixpoint_kind_of_rhs (Const (@{const_name lfp}, _) $ Abs _) = Lfp
  | fixpoint_kind_of_rhs (Const (@{const_name gfp}, _) $ Abs _) = Gfp
  | fixpoint_kind_of_rhs _ = NoFp
(* extended_context -> string * typ -> fixpoint_kind *)
fun fixpoint_kind_of_const ext_ctxt x =
  if is_built_in_const x then
    NoFp
  else
    fixpoint_kind_of_rhs (the (def_of_const ext_ctxt x)) handle Option => NoFp
(* extended_context -> (string * typ) -> bool *)
val is_inductive_pred = equal Lfp oo fixpoint_kind_of_const
(* extended_context -> (string * typ) -> bool *)
fun is_equational_fun
        (ext_ctxt as {thy, def_table, simp_table, psimp_table, ...}) x =
  is_inductive_pred ext_ctxt x
  orelse exists (fn table => not (null (def_props_for_const thy (!table) x)))
                [simp_table, psimp_table]

exception NoTriple

(* theory -> string * typ -> hol_term
   -> hol_term list * hol_term list * hol_term *)
fun triple_for_intro_rule thy x t =
  let
    val prems = Logic.strip_imp_prems t |> map (ObjectLogic.atomize_term thy)
    val concl = Logic.strip_imp_concl t |> ObjectLogic.atomize_term thy
    val (main, side) = List.partition (fn t => x mem Term.add_consts t []) prems
    (* hol_term -> bool *)
    val good_head = equal (Const x) o head_of
  in if forall good_head main then (side, main, concl) else raise NoTriple end

(* hol_term -> hol_term *)
val tuple_for_args = foldl1 HOLogic.mk_prod o snd o strip_comb

(* indexname * typ -> hol_term list -> hol_term -> hol_term -> hol_term *)
fun wf_constraint_for rel side concl main =
  let
    val core = HOLogic.mk_mem (HOLogic.mk_prod (tuple_for_args main,
                                                tuple_for_args concl), Var rel)
    val t = foldl HOLogic.mk_imp core side
    val vars = filter (not_equal rel) (Term.add_vars t [])
  in
    Library.foldl (fn (t', ((x, i), T)) =>
                      HOLogic.all_const T
                      $ Abs (x, T, abstract_over (Var ((x, i), T), t')))
                  (t, vars)
  end

(* indexname * typ -> hol_term list * hol_term list * hol_term -> hol_term *)
fun wf_constraint_for_triple rel (side, main, concl) =
  map (wf_constraint_for rel side concl) main |> foldl1 HOLogic.mk_conj

(* Proof.context -> Time.time -> thm -> (Proof.context -> tactic -> tactic)
   -> bool *)
fun terminates_by ctxt timeout goal tac =
  can (SINGLE (Classical.safe_tac (local_claset_of ctxt)) #> the
       #> SINGLE (timeout_tac timeout
                              (tac ctxt (auto_tac (local_clasimpset_of ctxt))))
       #> the #> Goal.finish) goal

(* Proof.context -> Time.time -> const_table -> string * typ -> bool *)
fun is_wellfounded_inductive_pred_no_cache ctxt timeout intro_table (s, T) =
  let
    val thy = ProofContext.theory_of ctxt
    val ts = def_props_for_const thy intro_table (s, T)
    val triples = map (triple_for_intro_rule thy (s, T)) ts
                  |> filter (not o null o #2)
  in
    if null triples then
      true
    else
      let
        val binders_T = foldl1 HOLogic.mk_prodT (binder_types T)
        val rel_T = HOLogic.mk_prodT (binders_T, binders_T) --> bool_T
        val j = foldl Int.max 0 (map maxidx_of_term ts) + 1
        val rel = (("R", j), rel_T)
        val prop = Const (@{const_name wf}, rel_T --> bool_T) $ Var rel
                   :: map (wf_constraint_for_triple rel) triples
                   |> foldl1 HOLogic.mk_conj |> HOLogic.mk_Trueprop
(*
        val _ = tracing ("***** Goal: " ^ Syntax.string_of_term ctxt prop)
*)
        val goal = cterm_of thy prop |> Goal.init
        val tacs = [LexicographicOrder.lexicographic_order_tac,
                    ScnpReconstruct.sizechange_tac]
        val res = silence (exists (terminates_by ctxt timeout goal)) tacs
      in res end
  end
  handle Empty => false
       | NoTriple => false

(* extended_context -> string * typ -> bool *)
fun is_wellfounded_inductive_pred
        {ctxt, inductive_mood, tac_timeout, intro_table, wfs, ...} x =
  case inductive_mood of
    Optimistic => true
  | Pessimistic => false
  | Realistic =>
    case AList.lookup op= (!wfs) x of
      SOME wf => wf
    | NONE =>
      let
        val wf = is_wellfounded_inductive_pred_no_cache ctxt tac_timeout
                                                        (!intro_table) x
        val _ = change wfs (cons (x, wf))
      in wf end

(* extended_context -> string * typ -> hol_term *)
fun better_founded_inductive_pred_const (ext_ctxt as {thy, simp_table, ...})
                                        (x as (s, T)) =
  let
    val iter_T = iterator_type_for_const x
    val x' as (s', _) = (wf_prefix ^ s, iter_T --> T)
    val surrogate = Const x' $ Const (@{const_name zero_class.zero}, iter_T)
  in
    (if not (is_equational_fun ext_ctxt x') then
       let
         val def = the (def_of_const ext_ctxt x)
         val j = maxidx_of_term def + 1
         val (outer, fp_app) = strip_abs def
         val outer_bounds = map Bound (length outer - 1 downto 0)
         val cur = Var (("i", j + 1), iter_T)
         val next = Const (@{const_name Suc}, iter_T --> iter_T) $ cur
         val rhs =
           case fp_app of
             Const _ $ t =>
             betapply (t, list_comb (Const x', next :: outer_bounds))
           | _ => raise BadHOLTerm ("inductive_fun_axiom", [fp_app])
         val (inner, naked_rhs) = strip_abs rhs
         val all = outer @ inner
         val bounds = map Bound (length all - 1 downto 0)
         val vars = map (fn (s, T) => Var ((s, j), T)) all
         val eq = HOLogic.mk_eq (list_comb (Const x', cur :: bounds), naked_rhs)
                  |> curry subst_bounds (rev vars)
       in
         simp_table := Symtab.update (s',
                           eq :: these (Symtab.lookup (!simp_table) s'))
                           (!simp_table)
       end
     else
       ());
    surrogate
  end

(* extended_context -> string * typ -> hol_term *)
fun raw_inductive_pred_axiom ext_ctxt x =
  let
    val def = the (def_of_const ext_ctxt x)
    val (outer, fp_app) = strip_abs def
    val outer_bounds = map Bound (length outer - 1 downto 0)
    val rhs = case fp_app of
                Const _ $ t => betapply (t, list_comb (Const x, outer_bounds))
              | _ => raise BadHOLTerm ("raw_inductive_pred_axiom", [fp_app])
    val (inner, naked_rhs) = strip_abs rhs
    val all = outer @ inner
    val bounds = map Bound (length all - 1 downto 0)
    val j = maxidx_of_term def + 1
    val vars = map (fn (s, T) => Var ((s, j), T)) all
    val eq = HOLogic.mk_eq (list_comb (Const x, bounds), naked_rhs)
  in subst_bounds (rev vars, eq) end

(* extended_context -> (string * typ) -> hol_term list *)
fun raw_equational_fun_axioms
        (ext_ctxt as {thy, simp_table, psimp_table, ...}) x =
  case def_props_for_const thy (!simp_table) x of
    [] => (case def_props_for_const thy (!psimp_table) x of
             [] => [raw_inductive_pred_axiom ext_ctxt x]
           | psimps => psimps)
  | simps => simps

val equational_fun_axioms = map extensionalize oo raw_equational_fun_axioms

(* string * typ -> int list *)
fun zombie_args_in_term x t =
  let
    (* hol_term list -> hol_term -> hol_term list -> hol_term list list *)
    fun fun_calls accum (Abs (_, _, t)) _ = fun_calls accum t []
      | fun_calls accum (t1 $ t2) arg_ts =
        fun_calls (fun_calls accum t2 []) t1 (t2 :: arg_ts)
      | fun_calls accum t arg_ts = (t = Const x ? cons arg_ts) accum
    (* hol_term list list -> hol_term list list -> hol_term list
       -> hol_term list list *)
    fun call_sets [] [] zs = [zs]
      | call_sets [] yss zs = zs :: call_sets yss [] []
      | call_sets ([] :: _) _ _ = []
      | call_sets ((x :: xs) :: xss) yss zs =
        call_sets xss (xs :: yss) (OrdList.insert TermOrd.term_ord x zs)
    val calls = fun_calls [] t []
    val sets = call_sets calls [] []
  in
    fold_rev (fn (j, set) => (length set = 1 andalso is_Var (hd set)) ? cons j)
             ((0 upto length sets - 1) ~~ sets) []
  end
(* string * typ -> hol_term list -> int list *)
fun zombie_args_in_terms x ts =
  fold1 (OrdList.inter int_ord) (map (zombie_args_in_term x) ts)

(* hol_term -> hol_term list *)
fun params_in_equation (@{const "==>"} $ _ $ t2) = params_in_equation t2
  | params_in_equation (@{const Trueprop} $ t1) = params_in_equation t1
  | params_in_equation (Const (@{const_name "op ="}, _) $ t1 $ _) =
    snd (strip_comb t1)
  | params_in_equation _ = []

(* string * typ -> string * typ -> int list -> hol_term list -> hol_term list
   -> hol_term -> hol_term *)
fun specialize_fun_axiom x x' fixed_js fixed_args bound_args t =
  let
    val fixed_params = filter_indices fixed_js (params_in_equation t)
    (* hol_term list -> hol_term -> hol_term *)
    fun aux args (Abs (s, T, t)) = list_comb (Abs (s, T, aux [] t), args)
      | aux args (t1 $ t2) = aux (aux [] t2 :: args) t1
      | aux args t =
        if t = Const x then
          list_comb (Const x', bound_args @ filter_out_indices fixed_js args)
        else
          let val j = find_index (equal t) fixed_params in
            list_comb (if j >= 0 then nth fixed_args j else t, args)
          end
  in aux [] t end

(* typ list -> hol_term -> bool *)
fun is_candidate_argument Ts t =
  let val bound_js = loose_bnos t in
    null bound_js
    orelse (is_higher_order_type (fastype_of1 (Ts, t))
            andalso forall (not o is_higher_order_type o nth Ts) bound_js)
  end

(* extended_context -> int -> hol_term -> hol_term *)
fun specialize_consts_in_hol_term
        (ext_ctxt as {special_depth, simp_table, special_funs, ...}) depth t =
  if depth > special_depth then
    t
  else
    let
      val black_list = if depth = 0 then [] else the_list (const_under_def t)
      (* hol_term list -> typ list -> hol_term -> hol_term *)
      fun aux args Ts (Const (x as (s, T))) =
          ((if not (x mem black_list) andalso not (null args)
               andalso not (String.isPrefix special_prefix s)
               andalso is_equational_fun ext_ctxt x then
              let
                val candidate_args =
                  fold_rev (fn (j, t) => is_candidate_argument Ts t ? cons j)
                           ((0 upto length args - 1) ~~ args) []
                val _ = not (null candidate_args) orelse raise Same
                val old_axs = equational_fun_axioms ext_ctxt x
                val zombie_params = zombie_args_in_terms x old_axs
                val fixed_js = OrdList.inter int_ord candidate_args
                                             zombie_params
                val _ = not (null fixed_js) orelse raise Same
                val fixed_args = filter_indices fixed_js args
                val bound_js = fold (fn t => fn js => add_loose_bnos (t, 0, js))
                                    fixed_args []
                               |> sort int_ord
                val variable_args = filter_out_indices fixed_js args
                val extra_args = map Bound bound_js @ variable_args
                val bound_Ts = filter_indices bound_js Ts
                val k = maxidx_of_term t + 1
                (* int -> hol_term *)
                fun free_for_bound_no j =
                  Var (("b" ^
                        nat_subscript (length bound_js
                                       - find_index (equal j) bound_js), k),
                       nth Ts j)
                val fixed_args_in_axiom =
                  map (curry subst_bounds
                             (map free_for_bound_no (0 upto length Ts - 1)))
                      fixed_args
              in
                case AList.lookup op= (!special_funs)
                                  (x, fixed_js, fixed_args_in_axiom) of
                  SOME x' => list_comb (Const x', extra_args)
                | NONE =>
                  let
                    val bound_args_in_axiom = map free_for_bound_no bound_js
                    val x' =
                      (special_prefix_for (length (!special_funs) + 1) ^ s,
                       bound_Ts @ filter_out_indices fixed_js (binder_types T)
                       ---> body_type T)
                    val new_axs =
                      map (specialize_fun_axiom x x' fixed_js
                               fixed_args_in_axiom bound_args_in_axiom) old_axs
                    val _ =
                      change special_funs
                             (cons ((x, fixed_js, fixed_args_in_axiom), x'))
                    val _ =
                      change simp_table (Symtab.update_new (fst x', new_axs))
                  in list_comb (Const x', extra_args) end
              end
            else
              raise Same)
           handle Same => list_comb (Const x, args))
        | aux args Ts (Abs (s, T, t)) =
          list_comb (Abs (s, T, aux [] (T :: Ts) t), args)
        | aux args Ts (t1 $ t2) = aux (aux [] Ts t2 :: args) Ts t1
        | aux args _ t = list_comb (t, args)
    in aux [] [] t end

val unfold_max_depth = 63
val axioms_max_depth = 63

(* extended_context -> hol_term -> hol_term *)
fun unfold_defs_in_hol_term (ext_ctxt as {thy, ...}) =
  let
    (* int -> typ list -> hol_term -> hol_term *)
    fun aux depth Ts t =
      case t of
        (t0 as Const (s as @{const_name Int.number_class.number_of},
                      T as Type ("fun", [_, T']))) $ t1 =>
        if is_integer_type T' then
          Const (numeral_prefix ^ Int.toString (HOLogic.dest_numeral t1), T')
          handle TERM _ =>
                 aux (depth + 1) Ts (the (def_of_const ext_ctxt (s, T)) $ t1)
        else
          betapply (t0, aux depth Ts t1)
      | Const x $ t1 $ t2 $ t3 => aux_const depth Ts x [t1, t2, t3]
      | Const x $ t1 $ t2 => aux_const depth Ts x [t1, t2]
      | Const x $ t1 => aux_const depth Ts x [t1]
      | Const x => aux_const depth Ts x []
      | Free _ => t
      | Var _ => t
      | Bound _ => t
      | Abs (s, T, body) => Abs (s, T, aux depth (T :: Ts) body)
      | t1 $ t2 => betapply (aux depth Ts t1, aux depth Ts t2)
    (* int -> typ list -> string * typ -> hol_term list -> hol_term *)
    and aux_const depth Ts (x as (s, _)) ts =
      if depth > unfold_max_depth then
        raise Limit ("unfold_defs_in_hol_term",
                     "too many nested definitions (" ^ Int.toString depth ^ ") \
                     \while expanding " ^ quote s)
      else
        let
          val (arity, const) =
            (case arity_of_built_in_const x of
              SOME n => (n, Const x)
            | NONE =>
              if is_inductive_pred ext_ctxt x
                 andalso not (is_wellfounded_inductive_pred ext_ctxt x) then
                (0, better_founded_inductive_pred_const ext_ctxt x)
              else if Refute.is_IDT_constructor thy x
                      orelse is_equational_fun ext_ctxt x then
                (0, Const x)
              else
                case def_of_const ext_ctxt x of
                  SOME def => (0, aux (depth + 1) Ts def)
                | NONE => (0, Const x))
          val const_of_ts =
            foldl (betapply o swap) const (map (aux (depth + 1) Ts) ts)
            |> Envir.beta_norm
        in eta_expand Ts const_of_ts (arity nat_minus length ts) end
  in aux 0 [] end

(* int -> hol_term -> hol_term *)
fun eval_axiom_for_hol_term j t =
  HOLogic.mk_eq (Const (eval_prefix ^ Int.toString j, fastype_of t), t)

(* extended_context -> hol_term -> hol_term list *)
fun axioms_for_hol_term (ext_ctxt as {thy, eval, def_table, nondef_table, ...})
                        t =
  let
    (* (int -> hol_term -> hol_term list -> hol_term list) -> int -> hol_term
       -> hol_term list -> hol_term list *)
    fun add_axiom depth t accum =
      let
        val t = unfold_defs_in_hol_term ext_ctxt t |> Refute.close_form
                |> specialize_consts_in_hol_term ext_ctxt depth
      in
        if member (op aconv) accum t then accum
        else add_axioms_for_term (depth + 1) t (t :: accum)
      end
    (* int -> hol_term -> hol_term list -> hol_term list *)
    and add_axioms_for_term depth t =
      case t of
        Const (x as (s, T)) =>
        (if is_built_in_const x then
           I
         else if depth > axioms_max_depth then
           raise Limit ("axioms_for_hol_term",
                        "too many nested axioms (" ^ Int.toString depth ^ ")")
         else if Refute.is_const_of_class thy x then
           let
             val class = Logic.class_of_const s
             val inclass = Logic.mk_inclass (TVar (("'a", 0), [class]), class)
             val ax1 = try (Refute.specialize_type thy x) inclass
             val ax2 = Option.map (Refute.specialize_type thy x o snd)
                                  (Refute.get_classdef thy class)
           in fold (add_axiom depth) (map_filter I [ax1, ax2]) end
         else if is_equational_fun ext_ctxt x then
           fold (add_axiom depth) (equational_fun_axioms ext_ctxt x)
         else if s = @{const_name The}
                 orelse s = @{const_name Hilbert_Choice.Eps} then
           raise Bad ("axioms_for_hol_term", "missing \"Nitpick\" theorems")
         else if is_ctor thy x then
           I
         else if is_abs_and_rep_fun thy x <> (false, false) then
           fold (fn table =>
                    fold (add_axiom depth)
                         (nondef_props_for_const thy table x))
                [!nondef_table,
                 Symtab.make [(s, Symtab.fold (fn (_, ts) => append ts)
                                              (!def_table) [])]]
         else
           fold (add_axiom depth)
                (nondef_props_for_const thy (!nondef_table) x))
        #> add_axioms_for_type depth T
      | Free (_, T) => add_axioms_for_type depth T
      | Var (_, T) => add_axioms_for_type depth T
      | Bound _ => I
      | Abs (_, T, t) => add_axioms_for_term depth t
                         #> add_axioms_for_type depth T
      | t1 $ t2 => add_axioms_for_term depth t1 #> add_axioms_for_term depth t2
    (* int -> typ -> hol_term list -> hol_term list *)
    and add_axioms_for_type depth T =
      case T of
        Type ("fun", Ts) => fold (add_axioms_for_type depth) Ts
      | Type ("*", Ts) => fold (add_axioms_for_type depth) Ts
      | @{typ prop} => I
      | @{typ bool} => I
      | @{typ unit} => I
      | @{typ nat} => I
      | @{typ int} => I
      | TFree (_, S) => add_axioms_for_sort depth T S
      | TVar (_, S) => add_axioms_for_sort depth T S
      | Type (s, Ts) =>
        fold (add_axioms_for_type depth) Ts
        #> (if s = @{const_name int} orelse is_datatype thy s then I
            else add_axiom_for_typedef depth T)
    (* int -> typ -> hol_term list -> hol_term list *)
    and add_axiom_for_typedef depth T =
      case Refute.get_typedef thy T of
        SOME (_, t) => add_axiom depth t
      | NONE => I
    (* int -> typ -> sort -> hol_term list -> hol_term list *)
    and add_axioms_for_sort depth T S =
      let
        val superclasses = S @ maps (Sign.super_classes thy) S |> distinct op=
        val class_axioms =
          maps (fn class =>
                   map Thm.prop_of (#axioms (AxClass.get_info thy class)
                                    handle ERROR _ => []))
               superclasses
        val monomorphic_class_axioms =
          map (fn t => case Term.add_tvars t [] of
                         [] => t
                       | [(x, S)] =>
                         Refute.monomorphic_term (Vartab.make [(x, (S, T))]) t
                       | _ => raise BadHOLTerm ("add_axioms_for_sort", [t]))
              class_axioms
      in fold (add_axiom depth) monomorphic_class_axioms end
    val eval_axioms = map2 eval_axiom_for_hol_term (0 upto length eval - 1) eval
  in add_axioms_for_term 1 t [] |> fold (add_axiom 1) eval_axioms end

(* (string * string list) list ref -> int -> bool -> hol_term -> hol_term *)
fun skolemize_hol_term skolems max_depth neg =
  let
    (* int list -> int list *)
    val incrs = map (curry op+ 1)
    (* string list -> typ list -> int list -> int -> bool -> hol_term
       -> hol_term *)
    fun aux ss Ts js depth neg t =
      let
        (* string -> typ -> string -> typ -> hol_term -> hol_term *)
        fun handle_quantifier quant_s quant_T abs_s abs_T t =
          if not (loose_bvar1 (t, 0)) then
            aux ss Ts js depth neg (incr_boundvars ~1 t)
          else if depth <= max_depth
                  andalso neg = (quant_s <> @{const_name Ex}) then
            let
              val j = length (!skolems) + 1
              val sko_s = skolem_prefix_for j ^ abs_s
              val _ = change skolems (cons (sko_s, ss))
              val sko_t = list_comb (Const (sko_s, rev Ts ---> abs_T),
                                     map Bound (rev js))
              val abs_t = Abs (abs_s, abs_T, aux ss Ts (incrs js) depth neg t)
            in
              if null js then betapply (abs_t, sko_t)
              else Const (@{const_name Let}, abs_T --> quant_T) $ sko_t $ abs_t
            end
          else
            Const (quant_s, quant_T)
            $ Abs (abs_s, abs_T,
                   (not (is_higher_order_type abs_T)
                    ? aux (abs_s :: ss) (abs_T :: Ts) (0 :: incrs js)
                          (depth + 1) neg) t)
      in
        case t of
          Const (s0 as @{const_name all}, T0) $ Abs (s1, T1, t1) =>
          handle_quantifier s0 T0 s1 T1 t1
        | Const (s0 as @{const_name All}, T0) $ Abs (s1, T1, t1) =>
          handle_quantifier s0 T0 s1 T1 t1
        | Const (s0 as @{const_name Ex}, T0) $ Abs (s1, T1, t1) =>
          handle_quantifier s0 T0 s1 T1 t1
        | @{const Trueprop} $ t1 =>
          @{const Trueprop} $ aux ss Ts js depth neg t1
        | @{const Not} $ t1 => @{const Not} $ aux ss Ts js depth (not neg) t1
        | @{const "==>"} $ t1 $ t2 =>
          @{const "==>"} $ aux ss Ts js depth (not neg) t1
          $ aux ss Ts js depth neg t2
        | @{const "op &"} $ t1 $ t2 =>
          @{const "op &"} $ aux ss Ts js depth neg t1
          $ aux ss Ts js depth neg t2
        | @{const "op |"} $ t1 $ t2 =>
          @{const "op |"} $ aux ss Ts js depth neg t1
          $ aux ss Ts js depth neg t2
        | @{const "op -->"} $ t1 $ t2 =>
          @{const "op -->"} $ aux ss Ts js depth (not neg) t1
          $ aux ss Ts js depth neg t2
        | (t0 as Const (@{const_name If}, _)) $ t1 $ t2 $ t3 =>
          t0 $ t1 $ aux ss Ts js depth neg t2 $ aux ss Ts js depth neg t3
        | (t0 as Const (@{const_name Let}, T0)) $ t1 $ t2 =>
          t0 $ t1 $ aux ss Ts js depth neg t2
        | Abs (s, T, t1) => Abs (s, T, aux ss Ts (incrs js) depth neg t1)
        | _ => t
      end
  in max_depth >= 0 ? aux [] [] [] 0 neg end

(* extended_context -> bool -> hol_term -> hol_term list * hol_term *)
fun preprocess_hol_term (ext_ctxt as {skolem_depth, skolems, ...}) negated =
  unfold_defs_in_hol_term ext_ctxt
  #> skolemize_hol_term skolems skolem_depth negated
  #> specialize_consts_in_hol_term ext_ctxt 0
  #> `(axioms_for_hol_term ext_ctxt)
  #> apfst (map unprefix_abs_vars_with_const_names)
  #> apsnd unprefix_abs_vars_with_const_names

(* Adapted from refute.ML *)
(* Proof.context -> hol_term list -> typ list *)
fun ground_types_in_hol_term ctxt ts =
  let
    val thy = ProofContext.theory_of ctxt
    (* typ -> typ list -> typ list *)
    fun add_type T =
      case T of
        Type ("fun", Ts) => fold add_type Ts
      | Type ("*", Ts) => fold add_type Ts
      | Type (@{type_name itself}, [T1]) => add_type T1
      | @{typ prop} => I
      | @{typ bool} => I
      | @{typ unit} => I
      | Type (s, Ts) =>
        (case DatatypePackage.get_datatype thy s of
           SOME {descr, index, ...} =>
           let
             val (_, typs, _) = the (AList.lookup op= descr index)
             val typ_assoc = typs ~~ Ts
             (* dtyp -> typ list -> typ list *)
             fun add_dtype U =
               let val T = Refute.typ_of_dtyp descr typ_assoc U in
                 case U of
                   DatatypeAux.DtTFree _ => add_type T
                 | DatatypeAux.DtType (_, Us) => add_type T #> fold add_dtype Us
                 | DatatypeAux.DtRec i =>
                   (fn accum =>
                       if T mem accum then
                         accum
                       else
                         let
                           val (_, dtyps, dconstrs) =
                             the (AList.lookup op= descr i)
                           val accum =
                             (not (is_pair_type T) ? insert op= T) accum
                             |> fold add_dtype dtyps
                         in
                           fold add_dtype (List.concat (map snd dconstrs)) accum
                         end)
               end
           in add_dtype (DatatypeAux.DtRec index) end
         | NONE => insert op= T #> fold add_type Ts)
      | TFree _ => insert op= T
      | TVar _ => insert op= T
  in foldl (OldTerm.it_term_types (uncurry add_type)) [] ts end

end;
