(*  Title:      Nitpick/Tools/nitpick_kodkod.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Kodkod problem generator part of Kodkod.
*)

signature NITPICK_KODKOD =
sig
  type dtype_spec = NitpickScope.dtype_spec
  type kodkod_ctors = NitpickPeephole.kodkod_ctors
  type term = NitpickTerm.term
  type ''a nfa_transition = Kodkod.rel_expr * ''a
  type ''a nfa_entry = ''a * ''a nfa_transition list
  type ''a nfa_table = ''a nfa_entry list

  structure NameTable : TABLE

  val univ_card : int -> int -> Kodkod.bound list -> Kodkod.formula -> int
  val kk_tuple : int -> int list -> Kodkod.tuple
  val bounds_for_built_in_rels_in_formula :
    bool -> int -> int -> int -> Kodkod.formula -> Kodkod.bound list
  val bound_for_plain_rel : term -> Kodkod.bound
  val bound_for_ctor_rel : dtype_spec list -> term -> Kodkod.bound
  val merge_bounds : Kodkod.bound list -> Kodkod.bound list -> Kodkod.bound list
  val is_only_liberally_satisfiable : Kodkod.raw_bound list -> bool
  val illiberalize_problem : Kodkod.problem -> Kodkod.problem
  val axiom_for_plain_rel : bool -> kodkod_ctors -> term -> Kodkod.formula
  val axiom_for_ctor_rel :
    bool -> dtype_spec list -> kodkod_ctors -> term -> Kodkod.formula
  val nfa_entry_for_dtype :
    kodkod_ctors -> dtype_spec list -> term NameTable.table -> dtype_spec
    -> typ nfa_entry
  val acyclicity_axiom : kodkod_ctors -> ''a nfa_table -> ''a -> Kodkod.formula
  val to_kodkod_formula :
    dtype_spec list -> bool -> bool -> bool -> bool -> kodkod_ctors -> term
    -> Kodkod.formula
end;

structure NitpickKodkod : NITPICK_KODKOD =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep
open NitpickTerm

type ''a nfa_transition = Kodkod.rel_expr * ''a
type ''a nfa_entry = ''a * ''a nfa_transition list
type ''a nfa_table = ''a nfa_entry list

(* int -> int -> Kodkod.bound list -> Kodkod.formula -> int *)
fun univ_card nat_card int_card bounds formula =
  let
    (* Kodkod.rel_expr -> int -> int *)
    fun rel_expr_func r k = Int.max (k, case r of
                                          Kodkod.Atom j => j + 1
                                        | Kodkod.AtomSeq (k', j0) => j0 + k'
                                        | _ => 0)
    (* Kodkod.tuple -> int -> int *)
    fun tuple_func t k = case t of
                           Kodkod.Tuple js => map (curry op+ 1) js
                                              |> foldl Int.max k
                         | _ => k
    (* Kodkod.tuple_set -> int -> int *)
    fun tuple_set_func ts k = Int.max (k,
                                       case ts of
                                         Kodkod.TupleAtomSeq (k', j0) => j0 + k'
                                       | _ => 0)
    val expr_F = {formula_func = K I, rel_expr_func = rel_expr_func,
                  int_expr_func = K I}
    val tuple_F = {tuple_func = tuple_func, tuple_set_func = tuple_set_func}
    val formula_card = fold (Kodkod.fold_bound expr_F tuple_F) bounds 1
                       |> Kodkod.fold_formula expr_F formula  in
    foldl Int.max 2 [nat_card, int_card, formula_card]
  end

(* string -> rep -> string *)
fun bound_comment nick R = nick ^ " : " ^ string_for_rep R

(* int -> int list -> Kodkod.tuple *)
fun kk_tuple univ_card js =
  Kodkod.TupleIndex (length js,
                     fold (fn j => fn accum => accum * univ_card + j) js 0)

(* Kodkod.formula -> Kodkod.n_ary_index list *)
fun built_in_rels_in_formula formula =
  let
    (* Kodkod.rel_expr -> Kodkod.n_ary_index list -> Kodkod.n_ary_index list *)
    fun rel_expr_func (Kodkod.Rel (n, j)) rels =
        (case AList.lookup op= (#rels initial_pool) n of
           SOME k => (j < k ? insert op= (n, j)) rels
         | NONE => rels)
      | rel_expr_func _ rels = rels
    val expr_F = {formula_func = K I, rel_expr_func = rel_expr_func,
                  int_expr_func = K I}
  in Kodkod.fold_formula expr_F formula [] end

(* int -> int -> (int -> int) -> Kodkod.tuple list *)
fun tabulate_unary_func univ_card k f =
  List.mapPartial
      (fn j1 => let val j2 = f j1 in
                  if j2 >= 0 then SOME (kk_tuple univ_card [j1, j2]) else NONE
                end)
      (0 upto k - 1)
(* int -> int -> (int * int -> int) -> Kodkod.tuple list *)
fun tabulate_binary_op univ_card k f =
  List.mapPartial
      (fn j =>
          let
            val j1 = j div k
            val j2 = j - j1 * k
            val j3 = f (j1, j2)
          in
            if j3 >= 0 then SOME (kk_tuple univ_card [j1, j2, j3]) else NONE
          end)
      (0 upto k * k - 1)
(* int -> int -> (int * int -> int) -> Kodkod.tuple list *)
fun tabulate_nat_binary_op univ_card k f =
  tabulate_binary_op univ_card k (atom_for_nat k o f)
fun tabulate_int_binary_op univ_card k f =
  tabulate_binary_op univ_card k
                     (atom_for_int k o f o pairself (int_for_atom k))

(* bool -> int -> int -> int -> Kodkod.rel_expr
   -> string * bool * Kodkod.tuple list *)
fun tabulate_built_in_rel weak univ_card nat_card int_card x =
  if Kodkod.Rel x = result_rel then
    ("result", weak, [Kodkod.Tuple [0]])
  else if Kodkod.Rel x = not3_rel then
    ("not3", false, tabulate_unary_func univ_card 2 (curry op- 1))
  else if Kodkod.Rel x = eq3_rel then
    ("eq3", false, tabulate_binary_op univ_card univ_card (bool_to_int o op=))
  else if Kodkod.Rel x = suc_rel then
    ("suc", false, tabulate_unary_func univ_card (univ_card - 1) (curry op+ 1))
  else if Kodkod.Rel x = nat_add_rel then
    ("nat_add", false, tabulate_nat_binary_op univ_card nat_card op+)
  else if Kodkod.Rel x = int_add_rel then
    ("int_add", false, tabulate_int_binary_op univ_card int_card op+)
  else if Kodkod.Rel x = nat_subtract_rel then
    ("nat_subtract", false,
     tabulate_binary_op univ_card nat_card (op nat_minus))
  else if Kodkod.Rel x = int_subtract_rel then
    ("int_subtract", false, tabulate_int_binary_op univ_card int_card op-)
  else if Kodkod.Rel x = nat_multiply_rel then
    ("nat_multiply", false, tabulate_nat_binary_op univ_card nat_card op*)
  else if Kodkod.Rel x = int_multiply_rel then
    ("int_multiply", false, tabulate_int_binary_op univ_card int_card op*)
  else if Kodkod.Rel x = int_less_rel then
    ("int_less", false,
     tabulate_int_binary_op univ_card int_card (bool_to_int o op<))
  else
    raise BadArg ("tabulate_built_in_rel", "unknown relation")

(* bool -> int -> int -> int -> Kodkod.rel_expr -> Kodkod.bound *)
fun bound_for_built_in_rel weak univ_card nat_card int_card x =
  let
    val (nick, opt, ts) =
      tabulate_built_in_rel weak univ_card nat_card int_card x
  in ([(x, nick)], (opt ? cons (Kodkod.TupleSet [])) [Kodkod.TupleSet ts]) end

(* bool -> int -> int -> int -> Kodkod.formula -> Kodkod.bound list *)
fun bounds_for_built_in_rels_in_formula weak univ_card nat_card int_card =
  map (bound_for_built_in_rel weak univ_card nat_card int_card)
  o built_in_rels_in_formula

(* rep -> Kodkod.tuple_set *)
val upper_bound_for_rep =
  foldl1 Kodkod.TupleProduct o map (fn k => Kodkod.TupleAtomSeq (k, 0))
  o atom_schema_of_rep

(* term -> Kodkod.bound *)
fun bound_for_plain_rel (FreeRel (x, _, R, nick)) =
    ([(x, bound_comment nick R)],
     [Kodkod.TupleSet [], upper_bound_for_rep R])
  | bound_for_plain_rel u = raise BadTerm ("bound_for_plain_rel", [u])

(* rep -> ctor_spec -> int -> Kodkod.tuple_set *)
fun upper_bound_for_self_rec_ctor_value R {typ, multiplicity, offset, ...}
                                        ran_j =
  let
    val binder_Rs = binder_reps R
    val binder_atom_schema = atom_schema_of_reps binder_Rs
    val binder_type_schema = type_schema_of_reps (binder_types typ) binder_Rs
    val body_T = body_type typ
    val end_offset = card_of_rep (body_rep R)
    val next_offset = offset + multiplicity
    val rec_tss = [(ran_j, 0), (end_offset - next_offset, next_offset)]
                  |> filter (not_equal 0 o fst)
                  |> map Kodkod.TupleAtomSeq
    val rec_ts = if null rec_tss then Kodkod.TupleAtomSeq (0, 0)
                                 else foldl1 Kodkod.TupleUnion rec_tss
    val dom_tss = map2 (fn k => fn T => if T = body_T then rec_ts
                                        else Kodkod.TupleAtomSeq (k, 0))
                       binder_atom_schema binder_type_schema
  in
    foldr Kodkod.TupleProduct (Kodkod.TupleSet [Kodkod.Tuple [ran_j]]) dom_tss
  end

(* rep -> ctor_spec -> Kodkod.tuple_set *)
fun upper_bound_for_self_rec_ctor R (ctor as {multiplicity, offset, ...}) =
  foldl1 Kodkod.TupleUnion (map (upper_bound_for_self_rec_ctor_value R ctor)
                                (index_seq offset multiplicity))

(* dtype_spec list -> term -> Kodkod.bound *)
fun bound_for_ctor_rel dtypes (FreeRel (x, T, R, nick)) =
    let
      val ctor as {multiplicity, total, self_rec, offset, ...} =
        ctor_spec dtypes (nick, T)
    in
      ([(x, bound_comment nick R)],
       if multiplicity = 0 then
         [Kodkod.TupleSet []]
       else
         let val binders_R = Struct (binder_reps R) in
           if total then
             let
               val dom_jss = all_combinations_for_rep binders_R
               val ran_js = index_seq offset multiplicity
             in
               if length dom_jss = length ran_js then
                 if body_rep R = Formula then
                   [Kodkod.TupleSet
                        (map (Kodkod.Tuple o fst)
                             (filter (equal 1 o snd) (dom_jss ~~ ran_js)))]
                 else
                   [Kodkod.TupleSet (map2 (fn dom_js => fn ran_j =>
                                              Kodkod.Tuple (dom_js @ [ran_j]))
                                          dom_jss ran_js)]
               else
                 raise BadArg ("bound_for_ctor_rel",
                               "bad cardinalities (" ^
                               Int.toString (length dom_jss) ^ " vs " ^
                               Int.toString (length ran_js) ^ ")")
             end
           else
             [Kodkod.TupleSet [],
              if self_rec andalso multiplicity > 0 then
                upper_bound_for_self_rec_ctor R ctor
              else
                Kodkod.TupleProduct (upper_bound_for_rep binders_R,
                    Kodkod.TupleAtomSeq (multiplicity, offset))]
         end)
    end
  | bound_for_ctor_rel _ u = raise BadTerm ("bound_for_ctor_rel", [u])

(* Kodkod.bound -> Kodkod.rel_expr *)
fun first_rel_of_bound (zs, _) = Kodkod.Rel (fst (hd zs))

(* Kodkod.bound list -> Kodkod.bound list -> Kodkod.bound list *)
fun merge_bounds [] cs = cs
  | merge_bounds (b :: bs) cs =
    let
      (* bound -> int *)
      fun arity (zs, _) = fst (fst (hd zs))
      (* bound -> bound list -> bound list -> bound list *)
      fun add_bound b [] ds = List.revAppend (ds, [b])
        | add_bound b (c :: cs) ds =
          if arity b = arity c andalso snd b = snd c
             andalso first_rel_of_bound b <> result_rel then
            List.revAppend (ds, (fst c @ fst b, snd c) :: cs)
          else
            add_bound b cs (c :: ds)
    in merge_bounds bs (add_bound b cs []) end

(* Kodkod.raw_bound list -> bool *)
fun is_only_liberally_satisfiable bounds =
  AList.lookup (op= o apsnd Kodkod.Rel) bounds result_rel = SOME []
(* Kodkod.bound -> Kodkod.bound *)
fun illiberalize_bound b = (first_rel_of_bound b = result_rel ? apsnd tl) b
(* Kodkod.formula -> Kodkod.formula *)
fun illiberalize_formula (Kodkod.And (f1, f2)) =
    s_and (illiberalize_formula f1) (illiberalize_formula f2)
  | illiberalize_formula (f as Kodkod.RelEq (Kodkod.None, r)) =
    if r = result_rel then Kodkod.False else f
  | illiberalize_formula f = f
(* Kodkod.problem -> Kodkod.problem *)
fun illiberalize_problem {comment, settings, univ_card, tuple_assigns, bounds,
                          int_bounds, expr_assigns, formula} =
  {comment = "illiberalized\n" ^ comment, settings = settings,
   univ_card = univ_card, tuple_assigns = tuple_assigns,
   bounds = map illiberalize_bound bounds, int_bounds = int_bounds,
   expr_assigns = expr_assigns, formula = illiberalize_formula formula}

(* int -> int -> Kodkod.rel_expr list *)
fun unary_var_seq j0 n = map (curry Kodkod.Var 1) (index_seq j0 n)

(* int list -> Kodkod.rel_expr *)
val singleton_from_combination = foldl1 Kodkod.Product o map Kodkod.Atom
(* rep -> Kodkod.rel_expr list *)
val all_singletons_for_rep =
  map singleton_from_combination o all_combinations_for_rep

(* Kodkod.rel_expr -> Kodkod.rel_expr list *)
fun unpack_products (Kodkod.Product (r1, r2)) =
    unpack_products r1 @ unpack_products r2
  | unpack_products r = [r]
fun unpack_joins (Kodkod.Join (r1, r2)) = unpack_joins r1 @ unpack_joins r2
  | unpack_joins r = [r]

(* rep -> Kodkod.rel_expr *)
val empty_rel_for_rep = empty_n_ary_rel o arity_of_rep
fun full_rel_for_rep R =
  case atom_schema_of_rep R of
    [] => raise BadRep ("full_rel_for_rep", [R])
  | schema => foldl1 Kodkod.Product (map (fn k => Kodkod.AtomSeq (k, 0)) schema)

(* int -> int list -> Kodkod.decl list *)
fun decls_for_atom_schema j0 schema =
  map2 (fn j => fn k => Kodkod.DeclOne ((1, j), Kodkod.AtomSeq (k, 0)))
       (index_seq j0 (length schema)) schema

(* The type constraint below is a workaround for a Poly/ML bug. *)

(* kodkod_ctors -> rep -> Kodkod.rel_expr -> Kodkod.formula *)
fun d_n_ary_function ({kk_all, kk_join, kk_lone, kk_one, ...} : kodkod_ctors)
                     R r =
  let val body_R = body_rep R in
    if is_lone_rep body_R then
      let
        val binder_schema = atom_schema_of_reps (binder_reps R)
        val body_schema = atom_schema_of_rep body_R
        val one = is_one_rep body_R
        val opt_x = case r of Kodkod.Rel x => SOME x | _ => NONE
      in
        if opt_x <> NONE andalso length binder_schema = 1
           andalso length body_schema = 1 then
          (if one then Kodkod.Function else Kodkod.Functional)
              (the opt_x, Kodkod.AtomSeq (hd binder_schema, 0),
               Kodkod.AtomSeq (hd body_schema, 0))
        else
          let
            val decls = decls_for_atom_schema ~1 binder_schema
            val vars = unary_var_seq ~1 (length binder_schema)
            val kk_xone = if one then kk_one else kk_lone
          in kk_all decls (kk_xone (fold kk_join vars r)) end
      end
    else
      Kodkod.True
  end

(* kodkod_ctors -> rep -> rel_expr -> formula *)
fun s_n_ary_function kk R (r as Kodkod.Rel _) =
    if not (is_opt_rep R) then
      if r = suc_rel then
        Kodkod.False
      else if r = nat_add_rel then
        bool_to_formula (card_of_rep (body_rep R) = 1)
      else if r = nat_multiply_rel then
        bool_to_formula (card_of_rep (body_rep R) <= 2)
      else
        d_n_ary_function kk R r
    else if r = nat_subtract_rel then
      Kodkod.True
    else
      d_n_ary_function kk R r
  | s_n_ary_function kk R r = d_n_ary_function kk R r

(* bool -> kodkod_ctors -> rep -> rel_expr -> formula *)
fun kk_n_ary_function peephole_optim =
  if peephole_optim then s_n_ary_function else d_n_ary_function

(* int -> kodkod_ctors -> (Kodkod.rel_expr -> Kodkod.rel_expr)
   -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun basic_rel_let j kk f r =
  if inline_rel_expr r then
    f r
  else
    let val n = Kodkod.arity_of_rel_expr r in
      #kk_rel_let kk [Kodkod.AssignRelReg ((n, j), r)]
                     (f (Kodkod.RelReg (n, j)))
    end

(* kodkod_ctors -> (Kodkod.rel_expr -> Kodkod.rel_expr) -> Kodkod.rel_expr
   -> Kodkod.rel_expr *)
val single_rel_let = basic_rel_let 0
(* kodkod_ctors -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr)
   -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun double_rel_let kk f r1 r2 =
  single_rel_let kk (fn r1 => basic_rel_let 1 kk (f r1) r2) r1
(* kodkod_ctors
   -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr)
   -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr
   -> Kodkod.rel_expr *)
fun triple_rel_let kk f r1 r2 r3 =
  double_rel_let kk (fn r1 => fn r2 => basic_rel_let 2 kk (f r1 r2) r3) r1 r2

(* kodkod_ctors -> Kodkod.rel_expr -> Kodkod.formula *)
fun atom_to_formula {kk_rel_eq, ...} r = kk_rel_eq r true_atom
(* kodkod_ctors -> Kodkod.formula -> Kodkod.rel_expr *)
fun formula_to_atom {kk_rel_if, ...} f = kk_rel_if f true_atom false_atom
(* kodkod_ctors -> rep -> Kodkod.formula -> Kodkod.rel_expr *)
fun formula_to_rel_expr kk R f =
  if R = Atom 2 orelse R = Opt (Atom 2) then formula_to_atom kk f
  else raise BadRep ("formula_to_rel_expr", [R])

(* kodkod_ctors -> Kodkod.formula -> Kodkod.formula -> Kodkod.formula *)
fun kk_notimplies {kk_not, kk_and, ...} f1 f2 = kk_and f1 (kk_not f2)

(* kodkod_ctors -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun kk_or3 kk =
  double_rel_let kk
      (fn r1 => fn r2 =>
          #kk_rel_if kk (#kk_subset kk true_atom (#kk_union kk r1 r2))
                        true_atom (#kk_intersect kk r1 r2))
fun kk_and3 kk =
  double_rel_let kk
      (fn r1 => fn r2 =>
          #kk_rel_if kk (#kk_subset kk false_atom (#kk_union kk r1 r2))
                        false_atom (#kk_intersect kk r1 r2))
fun kk_implies3 kk r1 r2 = kk_or3 kk (#kk_not3 kk r1) r2
fun kk_notimplies3 kk r1 r2 = kk_and3 kk r1 (#kk_not3 kk r2)

(* kodkod_cotrs -> int -> int -> Kodkod.rel_expr -> Kodkod.rel_expr list *)
fun unpack_vect_in_chunks {kk_project_seq, ...} chunk_arity num_chunks r =
  List.tabulate (num_chunks, fn j => kk_project_seq r (j * chunk_arity)
                                                    chunk_arity)

(* kodkod_ctors -> int -> int -> Kodkod.rel_expr -> Kodkod.rel_expr
   -> Kodkod.rel_expr *)
fun kk_n_fold_join {kk_join, ...} 1 _ r1 r2 = kk_join r1 r2
  | kk_n_fold_join kk arity1 res_R r1 r2 =
    let
      val unpacked_rs1 =
          if inline_rel_expr r1 then unpack_vect_in_chunks kk 1 arity1 r1
                                else unpack_products r1
    in
      if length unpacked_rs1 = arity1 then
        fold (#kk_join kk) unpacked_rs1 r2
      else
        #kk_project_seq kk
            (#kk_intersect kk (#kk_product kk r1 (full_rel_for_rep res_R)) r2)
            arity1 (arity_of_rep res_R)
    end

(* kodkod_ctors -> Kodkod.rel_expr -> Kodkod.rel_expr list
   -> Kodkod.rel_expr list -> Kodkod.rel_expr *)
fun kk_case_switch {kk_rel_let, kk_rel_if, kk_rel_eq, ...} r rs1 rs2 =
  if rs1 = rs2 then
    r
  else
    let
      val inline = (inline_rel_expr r orelse length rs1 <= 1)
      val r_arity = Kodkod.arity_of_rel_expr r
      val reg_r = if inline then r else Kodkod.RelReg (r_arity, 0)
      (* Kodkod.rel_expr list -> Kodkod.rel_expr list -> Kodkod.rel_expr *)
      fun aux [] [] = empty_n_ary_rel (Kodkod.arity_of_rel_expr (hd rs2))
        | aux (r1 :: rs1) (r2 :: rs2) =
          kk_rel_if (kk_rel_eq reg_r r1) r2 (aux rs1 rs2)
        | aux _ _ = raise BadArg ("kk_case_switch", "unequal cardinalities")
      val body_r = aux rs1 rs2
    in
      if inline then body_r
      else kk_rel_let [Kodkod.AssignRelReg ((r_arity, 0), r)] body_r
    end

(* kodkod_ctors -> int -> Kodkod.rel_expr -> Kodkod.formula list *)
fun unpack_formulas kk = map (atom_to_formula kk) oo unpack_vect_in_chunks kk 1

(* kodkod_ctors -> (Kodkod.formula -> Kodkod.formula -> Kodkod.formula)
   -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula *)
fun kk_vect_set_bool_op kk connective k r1 r2 =
  fold1 (#kk_and kk) (map2 connective (unpack_formulas kk k r1)
                           (unpack_formulas kk k r2))

(* kodkod_ctors -> (Kodkod.formula -> Kodkod.formula -> Kodkod.formula)
   -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun kk_vect_set_op kk connective k r1 r2 =
  fold1 (#kk_product kk) (map2 (formula_to_atom kk oo connective)
        (unpack_formulas kk k r1) (unpack_formulas kk k r2))

val lone_rep_fallback_max_card = 65536

(* kodkod_ctors -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun lone_rep_fallback kk new_R old_R r =
  let val card = card_of_rep old_R in
    if is_lone_rep old_R andalso is_lone_rep new_R
       andalso card = card_of_rep new_R then
      if card > lone_rep_fallback_max_card then
        raise Limit ("lone_rep_fallback",
                     "too high cardinality (" ^ Int.toString card ^ ")")
      else
        kk_case_switch kk r (all_singletons_for_rep old_R)
                       (all_singletons_for_rep new_R)
    else
      raise BadRep ("lone_rep_fallback", [old_R, new_R])
  end
(* kodkod_ctors -> int -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_to_atom kk k old_R r =
  case old_R of
    Atom _ => r
  | Func (R1, R2) =>
    let
      val dom_card = card_of_rep R1
      val R2' = Atom (card_of_rep R2)
    in
      rel_expr_to_atom kk k (Vect (dom_card, R2'))
                       (rel_expr_to_vect kk dom_card R2' old_R r)
    end
  | Opt _ => raise BadRep ("rel_expr_to_atom", [old_R])
  | _ => lone_rep_fallback kk (Atom k) old_R r
(* kodkod_ctors -> rep list -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_to_struct kk Rs old_R r =
  case old_R of
    Atom _ => lone_rep_fallback kk (Struct Rs) old_R r
  | Struct Rs' =>
    let
      val Rs = filter (not_equal Unit) Rs
      val Rs' = filter (not_equal Unit) Rs'
    in
      if Rs' = Rs then
        r
      else if map card_of_rep Rs' = map card_of_rep Rs then
        let
          val old_arities = map arity_of_rep Rs'
          val old_offsets = offset_list old_arities
          val old_rs = map2 (#kk_project_seq kk r) old_offsets old_arities
        in fold1 (#kk_product kk) (map3 (rel_expr_to_rep kk) Rs Rs' old_rs) end
      else
        lone_rep_fallback kk (Struct Rs) old_R r
    end
  | _ => raise BadRep ("rel_expr_to_struct", [old_R])
(* kodkod_ctors -> int -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_to_vect kk k R old_R r =
  case old_R of
    Atom _ => lone_rep_fallback kk (Vect (k, R)) old_R r
  | Vect (k', R') =>
    if k = k' andalso R = R' then r
    else lone_rep_fallback kk (Vect (k, R)) old_R r
  | Func (R1, Formula) =>
    if k = card_of_rep R1 then
      fold1 (#kk_product kk)
            (map (fn arg_r =>
                     formula_to_rel_expr kk R (#kk_subset kk arg_r r))
                 (all_singletons_for_rep R1))
    else
      raise BadRep ("to_vect", [old_R])
  | Func (Unit, R2) => rel_expr_to_rep kk R R2 r
  | Func (R1, R2) =>
    fold1 (#kk_product kk)
          (map (fn arg_r =>
                   rel_expr_to_rep kk R R2
                       (kk_n_fold_join kk (arity_of_rep R1) R2 arg_r r))
               (all_singletons_for_rep R1))
  | _ => raise BadRep ("rel_expr_to_vect", [old_R])
(* kodkod_ctors -> rep -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_to_func_no_opt kk R1 R2 (Atom k) r =
    let
      val dom_card = card_of_rep R1
      val R2' = Atom (card_of_rep R2)
    in
      rel_expr_to_func_no_opt kk R1 R2 (Vect (dom_card, R2'))
                              (rel_expr_to_vect kk dom_card R2' (Atom k) r)
    end
  | rel_expr_to_func_no_opt kk Unit R2 old_R r =
    (case old_R of
       Vect (k, R') => rel_expr_to_rep kk R2 R' r
     | Func (Unit, R2') => rel_expr_to_rep kk R2 R2' r
     | Func (Atom 1, Formula) => formula_to_atom kk (#kk_some kk r)
     | Func (R1', R2') =>
       rel_expr_to_rep kk R2 R2' (#kk_project_seq kk r (arity_of_rep R1')
                                                  (arity_of_rep R2'))
     | _ => raise BadRep ("rel_expr_to_func_no_opt", [old_R, Func (Unit, R2)]))
  | rel_expr_to_func_no_opt kk R1 Formula old_R r =
    (case old_R of
       Vect (k, Atom 2) =>
       let
         val args_rs = all_singletons_for_rep R1
         val vals_rs = unpack_vect_in_chunks kk 1 k r
         (* Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
         fun empty_or_singleton_set_for arg_r val_r =
             #kk_join kk val_r (#kk_product kk true_atom arg_r)
       in
         fold1 (#kk_union kk) (map2 empty_or_singleton_set_for args_rs vals_rs)
       end
     | Func (R1', Formula) =>
       if R1 = R1' then
         r
       else
         let
           val schema = atom_schema_of_rep R1
           val r1 = fold1 (#kk_product kk) (unary_var_seq ~1 (length schema))
                    |> rel_expr_to_rep kk R1' R1
         in
           #kk_comprehension kk (decls_for_atom_schema ~1 schema)
                                (#kk_subset kk r1 r)
         end
     | Func (Unit, Atom 2) =>
       #kk_rel_if kk (#kk_rel_eq kk r true_atom) atom_0 Kodkod.None
     | Func (R1', Atom 2) =>
       rel_expr_to_func_no_opt kk R1 Formula (Func (R1', Formula))
                               (#kk_join kk r true_atom)
     | _ => raise BadRep ("rel_expr_to_func_no_opt",
                          [old_R, Func (R1, Formula)]))
  | rel_expr_to_func_no_opt kk R1 R2 old_R r =
    (case old_R of
       Vect (k, R) =>
       let
         val args_rs = all_singletons_for_rep R1
         val vals_rs = unpack_vect_in_chunks kk (arity_of_rep R) k r
                       |> map (rel_expr_to_rep kk R2 R)
       in fold1 (#kk_union kk) (map2 (#kk_product kk) args_rs vals_rs) end
     | Func (R1', Formula) =>
       if R2 = Atom 2 then
         let val schema = atom_schema_of_rep R1 in
           if length schema = 1 then
             #kk_override kk (#kk_product kk (Kodkod.AtomSeq (hd schema, 0))
                                             false_atom)
                             (#kk_product kk r true_atom)
           else
             let
               val r1 = fold1 (#kk_product kk)
                              (unary_var_seq ~1 (length schema))
                        |> rel_expr_to_rep kk R1' R1
               val r2 = Kodkod.Var (1, ~(length schema) - 1)
               val r3 = formula_to_atom kk (#kk_subset kk r1 r)
             in
               #kk_comprehension kk (decls_for_atom_schema ~1 (schema @ [2]))
                                 (#kk_rel_eq kk r2 r3)
             end
           end
         else
           raise BadRep ("rel_expr_to_func_no_opt", [old_R, Func (R1, R2)])
     | Func (Unit, R2') =>
       rel_expr_to_func_no_opt kk R1 R2 (Func (Atom 1, R2'))
                               (#kk_product kk atom_0 r)
     | Func (R1', R2') =>
       if R1 = R1' andalso R2 = R2' then
         r
       else
         let
           val dom_schema = atom_schema_of_rep R1
           val ran_schema = atom_schema_of_rep R2
           val dom_prod = fold1 (#kk_product kk)
                                (unary_var_seq ~1 (length dom_schema))
                          |> rel_expr_to_rep kk R1' R1
           val ran_prod = fold1 (#kk_product kk)
                                (unary_var_seq (~(length dom_schema) - 1)
                                               (length ran_schema))
                          |> rel_expr_to_rep kk R2' R2
           val app = kk_n_fold_join kk (arity_of_rep R1') R2' dom_prod r
         in
           #kk_comprehension kk (decls_for_atom_schema ~1
                                    (dom_schema @ ran_schema))
                                (#kk_subset kk ran_prod app)
         end
     | _ => raise BadRep ("rel_expr_to_func_no_opt", [old_R, Func (R1, R2)]))
(* kodkod_ctors -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_to_rep kk new_R old_R r =
  let
    val unopt_old_R = unopt_rep old_R
    val unopt_new_R = unopt_rep new_R
  in
    if unopt_old_R <> old_R andalso unopt_new_R = new_R then
      raise BadRep ("rel_expr_to_rep", [old_R, new_R])
    else if unopt_new_R = unopt_old_R then
      r
    else
      (case unopt_new_R of
         Atom k => rel_expr_to_atom kk k
       | Struct Rs => rel_expr_to_struct kk Rs
       | Vect (k, R') => rel_expr_to_vect kk k R'
       | Func (R1, R2) => rel_expr_to_func_no_opt kk R1 R2
       | _ => raise BadRep ("rel_expr_to_rep", [old_R, new_R])) unopt_old_R r
  end
(* kodkod_ctors -> rep -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_to_func kk R1 R2 = rel_expr_to_rep kk (Func (R1, R2))

(* bool -> kodkod_ctors -> term -> Kodkod.formula *)
fun axiom_for_plain_rel peephole_optim kk (FreeRel (x, _, R as Func _, _)) =
    kk_n_ary_function peephole_optim kk R (Kodkod.Rel x)
  | axiom_for_plain_rel _ {kk_lone, kk_one, ...} (FreeRel (x, _, R, _)) =
    if is_one_rep R then kk_one (Kodkod.Rel x)
    else if is_lone_rep R andalso card_of_rep R > 1 then kk_lone (Kodkod.Rel x)
    else Kodkod.True
  | axiom_for_plain_rel _ _ u = raise BadTerm ("axiom_for_plain_rel", [u])

(* bool -> dtype_spec list -> kodkod_ctors -> term -> Kodkod.formula *)
fun axiom_for_ctor_rel peephole_optim dtypes kk (FreeRel (x, T, R, nick)) =
    let val {multiplicity, total, offset, ...} = ctor_spec dtypes (nick, T) in
      if total orelse multiplicity = 0 then
        Kodkod.True
      else
        let
          val decl_axiom = kk_n_ary_function peephole_optim kk R (Kodkod.Rel x)
          val distinct_and_onto_axiom =
            #kk_all kk
                [Kodkod.DeclOne ((1, 0), Kodkod.AtomSeq (multiplicity, offset))]
                (#kk_one kk (#kk_join kk (Kodkod.Rel x) (Kodkod.Var (1, 0))))
        in #kk_and kk decl_axiom distinct_and_onto_axiom end
    end
  | axiom_for_ctor_rel _ _ _ u = raise BadTerm ("axiom_for_ctor_rel", [u])

(* kodkod_ctors -> dtype_spec list -> term NameTable.table -> ctor_spec
   -> ''a nfa_transition list *)
fun nfa_transitions_for_ctor {kk_project, ...} dtypes table {name, typ, ...} =
  let
    val u = lookup_name table (ConstName (name, typ, Any))
    val (r, R, arity) =
      case u of
        FreeRel ((n, j), _, R, _) => (Kodkod.Rel (n, j), R, n)
      | _ => raise BadTerm ("nfa_transitions_for_ctor", [u])
    val binder_Rs = binder_reps R
    val binder_atom_schema = atom_schema_of_reps binder_Rs
    val binder_type_schema = type_schema_of_reps (binder_types typ) binder_Rs
    val body_T = body_type typ
  in
    List.mapPartial
        (fn (T, j) =>
            if exists (equal T o #typ) dtypes then
              SOME (kk_project r [Kodkod.Num j, Kodkod.Num (arity - 1)], T)
            else
              NONE)
        (binder_type_schema ~~ (0 upto length binder_type_schema - 1))
  end
  handle BadTerm _ => [] (* for Unit *)

(* kodkod_ctors -> dtype_spec list -> term NameTable.table -> dtype_spec
   -> typ * typ nfa_transition list *)
fun nfa_entry_for_dtype kk dtypes rel_table {typ, ctors, ...} =
  (typ, List.concat (map (nfa_transitions_for_ctor kk dtypes rel_table) ctors))

val empty_rel = Kodkod.Product (Kodkod.None, Kodkod.None)

(* nfa_table -> ''a list *)
fun all_states table = map fst table

(* nfa_table -> ''a -> ''a -> Kodkod.rel_expr list *)
fun direct_path_rel_exprs table start final =
  case AList.lookup op= table final of
    SOME trans => map fst (filter (equal start o snd) trans)
  | NONE => []
(* kodkod_ctors -> nfa_table -> ''a list -> ''a -> ''a -> Kodkod.rel_expr *)
and any_path_rel_expr {kk_union, ...} table [] start final =
    fold kk_union (direct_path_rel_exprs table start final)
         (if start = final then Kodkod.Iden else empty_rel)
  | any_path_rel_expr kk table (q :: qs) start final =
    #kk_union kk (any_path_rel_expr kk table qs start final)
                 (knot_path_rel_expr kk table qs start q final)
(* kodkod_ctors -> nfa_table -> ''a list -> ''a -> ''a -> ''a
   -> Kodkod.rel_expr *)
and knot_path_rel_expr kk table qs start knot final =
  #kk_join kk (any_path_rel_expr kk table qs start knot)
              (#kk_join kk (#kk_reflexive_closure kk
                                (loop_path_rel_expr kk table qs knot))
                           (any_path_rel_expr kk table qs knot final))
(* kodkod_ctors -> nfa_table -> ''a list -> ''a -> Kodkod.rel_expr *)
and loop_path_rel_expr {kk_union, ...} table [] start =
    fold kk_union (direct_path_rel_exprs table start start) empty_rel
  | loop_path_rel_expr kk table (q :: qs) start =
    if start = q then
      #kk_closure kk (loop_path_rel_expr kk table qs start)
    else
      #kk_union kk (loop_path_rel_expr kk table qs start)
                   (knot_path_rel_expr kk table qs start q start)

(* kodkod_ctors -> ''a nfa_table -> ''a -> Kodkod.formula *)
fun acyclicity_axiom kk table start =
  #kk_no kk (#kk_intersect kk
                 (loop_path_rel_expr kk table (all_states table) start)
                 Kodkod.Iden)

(* dtype_spec list -> bool -> bool -> bool -> kodkod_ctors -> term
   -> Kodkod.formula *)
fun to_kodkod_formula dtypes peephole_optim use_result_rel liberal negate kk u =
  let
    val {kk_all, kk_exist, kk_formula_let, kk_formula_if, kk_or, kk_not, kk_iff,
         kk_implies, kk_and, kk_subset, kk_rel_eq, kk_no, kk_some, kk_rel_let,
         kk_rel_if, kk_union, kk_difference, kk_intersect, kk_product, kk_join,
         kk_closure, kk_comprehension, kk_project_seq, kk_not3, kk_eq3,
         kk_nat_less, kk_int_less, ...} = kk

    (* term -> Kodkod.formula *)
    fun to_f u =
      case rep_of u of
        Formula =>
        (case u of
           Cst (False, _, _) => Kodkod.False
         | Cst (True, _, _) => Kodkod.True
         | Op1 (Not, _, _, u1) => kk_not (to_f u1)
         | Op1 (Finite, _, _, _) => Kodkod.True
         | Op1 (Cast, _, _, u1) => to_f u1
         | Op2 (All, _, _, u1, u2) => kk_all (untuple to_decl u1) (to_f u2)
         | Op2 (Exist, _, _, u1, u2) => kk_exist (untuple to_decl u1) (to_f u2)
         | Op2 (Or, _, _, u1, u2) => kk_or (to_f u1) (to_f u2)
         | Op2 (Implies, _, _, u1, u2) => kk_implies (to_f u1) (to_f u2)
         | Op2 (And, _, _, u1, u2) => kk_and (to_f u1) (to_f u2)
         | Op2 (StrictEq, _, _, u1, u2) =>
           (case min_rep (rep_of u1) (rep_of u2) of
              Unit => Kodkod.True
            | Formula => kk_iff (to_f u1) (to_f u2)
            | min_R =>
              let
                (* term -> term list *)
                fun args (Op2 (Apply, _, _, u1, u2)) = u2 :: args u1
                  | args _ = []
                val opt_arg_us = filter (is_opt_rep o rep_of) (args u1)
              in
                fold (kk_or o (kk_no o to_r)) opt_arg_us
                     (kk_rel_eq (to_rep min_R u1) (to_rep min_R u2))
              end)
         | Op2 (Eq, _, _, u1, u2) =>
           (case min_rep (rep_of u1) (rep_of u2) of
              Unit => Kodkod.True
            | Formula => kk_iff (to_f u1) (to_f u2)
            | min_R =>
              let
                val r1 = to_rep min_R u1
                val r2 = to_rep min_R u2
              in
                if is_one_rep min_R then
                  let
                    val rs1 = unpack_products r1
                    val rs2 = unpack_products r2
                  in
                    if length rs1 = length rs2
                       andalso map Kodkod.arity_of_rel_expr rs1
                               = map Kodkod.arity_of_rel_expr rs2 then
                      fold1 kk_and (map2 kk_rel_eq rs1 rs2)
                    else
                      kk_rel_eq r1 r2
                  end
                else
                  kk_rel_eq r1 r2
              end)
         | Op2 (Subset, _, _, u1, u2) =>
           to_set_bool_op kk_implies kk_subset u1 u2
         | Op2 (Apply, T, _, u1, u2) =>
           (case rep_of u1 of
              Func (R, Formula) => kk_subset (to_rep R u2) (to_r u1)
            | R => to_f (Op2 (Apply, T, Atom 2, u1, u2)))
         | Op3 (Let, _, _, u1, u2, u3) =>
           kk_formula_let [to_expr_assign u1 u2] (to_f u3)
         | Op3 (If, _, _, u1, u2, u3) =>
           kk_formula_if (to_f u1) (to_f u2) (to_f u3)
         | FormulaReg (j, _, _) => Kodkod.FormulaReg j
         | _ => raise BadTerm ("to_f", [u]))
      | Atom 2 => atom_to_formula kk (to_r u)
      | R => raise BadRep ("to_f", [R])
    (* term -> Kodkod.rel_expr *)
    and to_r u =
      case u of
        Cst (Unity, _, Unit) => raise BadTerm ("to_r", [u])
      | Cst (False, _, Atom 2) => false_atom
      | Cst (True, _, Atom 2) => true_atom
      | Cst (None, _, Func (R, Formula)) => empty_rel_for_rep R
      | Cst (Univ, _, Func (R, Formula)) => full_rel_for_rep R
      | Cst (Iden, T, Func (Struct [R1, R2], Formula)) =>
        if R1 = R2 andalso arity_of_rep R1 = 1 then
          kk_intersect Kodkod.Iden (kk_product (full_rel_for_rep R1)
                                               Kodkod.Univ)
        else
          let
            val schema1 = atom_schema_of_rep R1
            val schema2 = atom_schema_of_rep R2
            val arity1 = length schema1
            val arity2 = length schema2
            val r1 = fold1 kk_product (unary_var_seq 0 arity1)
            val r2 = fold1 kk_product (unary_var_seq arity1 arity2)
            val min_R = min_rep R1 R2
          in
            kk_comprehension
                (decls_for_atom_schema 0 (schema1 @ schema2))
                (kk_rel_eq (rel_expr_to_rep kk min_R R1 r1)
                           (rel_expr_to_rep kk min_R R2 r2))
          end
      | Cst (Iden, T, Func (Atom 1, Formula)) => atom_0
      | Cst (Num j, @{typ int}, R) =>
         (case atom_for_int (card_of_rep R) j of
            ~1 => if is_opt_rep R then Kodkod.None
                  else raise BadTerm ("to_r", [u])
          | j' => Kodkod.Atom j')
      | Cst (Num j, _, R) =>
         if j < card_of_rep R then Kodkod.Atom j
         else if is_opt_rep R then Kodkod.None
         else raise BadTerm ("to_r", [u])
      | Cst (Biggie, _, Opt R) => empty_rel_for_rep R
      | Cst (Suc, _, Func (Unit, _)) => Kodkod.None
      | Cst (Suc, _, Func (Atom k, _)) =>
        kk_intersect suc_rel (kk_product Kodkod.Univ (Kodkod.AtomSeq (k, 0)))
      | Cst (Add, Type ("fun", [@{typ nat}, _]), _) => nat_add_rel
      | Cst (Add, Type ("fun", [@{typ int}, _]), _) => int_add_rel
      | Cst (Subtract, Type ("fun", [@{typ nat}, _]), _) => nat_subtract_rel
      | Cst (Subtract, Type ("fun", [@{typ int}, _]), _) => int_subtract_rel
      | Cst (Multiply, Type ("fun", [@{typ nat}, _]), _) => nat_multiply_rel
      | Cst (Multiply, Type ("fun", [@{typ int}, _]), _) => int_multiply_rel
      | Cst (NatToInt, _, Func (_, Atom _)) => Kodkod.Iden
      | Cst (NatToInt, _, Func (_, Opt (Atom k))) =>
        kk_intersect Kodkod.Iden
            (kk_product (Kodkod.AtomSeq (max_int_for_card k + 1, 0))
                        Kodkod.Univ)
      | Cst (IntToNat, _, Func (Atom k1, R2)) =>
        let
          val abs_card = max_int_for_card k1 + 1
          val nat_card = card_of_rep R2
          val overlap = Int.min (nat_card, abs_card)
        in
          kk_union
              (kk_product (Kodkod.AtomSeq (k1 - abs_card, abs_card)) atom_0)
              (kk_intersect Kodkod.Iden
                   (kk_product (Kodkod.AtomSeq (overlap, 0)) Kodkod.Univ))
        end
      | Op1 (Not, _, R as Opt (Atom 2), u1) => kk_not3 (to_rep R u1)
      | Op1 (Finite, _, Opt (Atom 2), _) => Kodkod.None
      | Op1 (Card, _, R, u1) =>
        let val R1 = rep_of u1 in
          if is_opt_rep R1 then
            Kodkod.None
          else
            let
              val k = exact_log 2 (card_of_rep R1)
              (* Kodkod.rel_expr -> int -> Kodkod.rel_expr *)
              fun count_elem r j =
                kk_rel_if (kk_subset (Kodkod.Atom j) r) suc_rel Kodkod.Iden
            in
              single_rel_let kk
                  (fn r => fold (fn j => fn accum =>
                                    kk_join accum (count_elem r j))
                                (0 upto k - 1) atom_0)
                  (to_func (Atom k) Formula u1)
              |> is_opt_rep R ? kk_intersect (Kodkod.AtomSeq (card_of_rep R, 0))
            end
        end
      | Op1 (Closure, _, R as Func (Atom 1, Formula), u1) => to_rep R u1
      | Op1 (Closure, _, R, u1) =>
        if is_opt_rep R then
          let
            val R' = rep_to_binary_rel_rep (unopt_rep (rep_of u1))
            val R'' = Func (domain_rep R', Opt (Atom 2))
            val r1 = kk_join (to_rep R'' u1) true_atom
          in rel_expr_to_rep kk R R'' (kk_product (kk_closure r1) true_atom) end
        else
          let val R' = rep_to_binary_rel_rep (rep_of u1) in
            rel_expr_to_rep kk R R' (kk_closure (to_rep R' u1))
          end
      | Op1 (SingletonSet, _, Func (R1, Opt (Atom 2)), Cst (Biggie, _, _)) =>
        if R1 = Unit then false_atom
        else kk_product (full_rel_for_rep R1) false_atom
      | Op1 (SingletonSet, _, R, u1) =>
        (case R of
           Func (R1, Formula) => to_rep R1 u1
         | Func (Unit, Opt (Atom 2)) => to_opt (Atom 1) u1
         | Func (R1, R2 as Opt (Atom 2)) =>
           single_rel_let kk
               (fn r => kk_rel_if (kk_no r) (empty_rel_for_rep R)
                                  (rel_expr_to_func kk R1 (Atom 2)
                                                    (Func (R1, Formula)) r))
               (to_opt R1 u1)
         | _ => raise BadRep ("to_r (SingletonSet)", [R]))
      | Op1 (First, _, R, Tuple (_, _, [u11, _])) => to_rep R u11
      | Op1 (First, _, R, u1) =>
        (case rep_of u1 of
           Opt (Atom 1) => to_rep R u1
         | Atom k' =>
           let val k = card_of_rep R in
             rel_expr_to_rep kk R (Atom k)
                 (kk_join (to_rep (Struct [Atom k, Atom (k' div k)]) u1)
                          Kodkod.Univ)
           end
         | Struct [R11, _] => to_project R R11 (to_r u1) 0
         | Opt (Struct [R11, _]) => to_project R (Opt R11) (to_r u1) 0
         | R1 => raise BadRep ("to_r (First)", [R1]))
      | Op1 (Second, _, R, Tuple (_, _, [_, u12])) => to_rep R u12
      | Op1 (Second, _, R, u1) =>
        (case rep_of u1 of
           Opt (Atom 1) => to_rep R u1
         | Atom k' =>
           let val k = card_of_rep R in
             rel_expr_to_rep kk R (Atom k)
                 (kk_join Kodkod.Univ
                          (to_rep (Struct [Atom (k' div k), Atom k]) u1))
           end
         | Struct [R11, R12] => to_project R R12 (to_r u1) (arity_of_rep R11)
         | Opt (Struct [R11, R12]) => to_project R (Opt R12) (to_r u1)
                                                 (arity_of_rep R11)
         | R1 => raise BadRep ("to_r (Second)", [R1]))
      | Op1 (Cast, _, R, u1) =>
        if rep_of u1 = Formula andalso (R = Atom 2 orelse R = Opt (Atom 2)) then
          formula_to_atom kk (to_f u1)
        else
          rel_expr_to_rep kk R (rep_of u1) (to_r u1)
      | Op2 (All, T, R as Opt (Atom 2), u1, u2) =>
        to_r (Op1 (Not, T, R,
                   Op2 (Exist, T, R, u1, Op1 (Not, T, rep_of u2, u2))))
      | Op2 (Exist, T, Opt (Atom 2), u1, u2) =>
        let val rs1 = untuple to_decl u1 in
          if not (is_opt_rep (rep_of u2)) then
            kk_rel_if (kk_exist rs1 (to_f u2)) true_atom Kodkod.None
          else
            single_rel_let kk
                (fn r2 =>
                    kk_rel_if (kk_exist rs1 (kk_rel_eq r2 true_atom))
                              true_atom
                              (kk_rel_if (kk_all rs1 (kk_rel_eq r2 false_atom))
                                         false_atom Kodkod.None))
                (to_opt (Atom 2) u2)
        end
      | Op2 (Or, _, _, u1, u2) =>
        (case (rep_of u1, rep_of u2) of
           (Opt (Atom 2), Opt (Atom 2)) => kk_or3 kk (to_r u1) (to_r u2)
         | (Opt (Atom 2), _) => kk_rel_if (to_f u2) true_atom (to_r u1)
         | (_, _) => kk_rel_if (to_f u1) true_atom (to_r u2))
      | Op2 (Implies, _, _, u1, u2) =>
        (case (rep_of u1, rep_of u2) of
           (Opt (Atom 2), Opt (Atom 2)) => kk_implies3 kk (to_r u1) (to_r u2)
         | (Opt (Atom 2), _) => kk_rel_if (to_f u2) true_atom
                                          (kk_not3 (to_r u1))
         | (_, _) => kk_rel_if (to_f u1) (to_r u2) true_atom)
      | Op2 (And, _, _, u1, u2) =>
        (case (rep_of u1, rep_of u2) of
           (Opt (Atom 2), Opt (Atom 2)) => kk_and3 kk (to_r u1) (to_r u2)
         | (Opt (Atom 2), _) => kk_rel_if (to_f u2) (to_r u1) false_atom
         | (_, _) => kk_rel_if (to_f u1) (to_r u2) false_atom)
      | Op2 (Less, _, _, u1, u2) =>
        if type_of u1 = nat_T then
          if is_Biggie u1 then to_compare_with_biggie u2 false_atom
          else if is_Biggie u2 then to_compare_with_biggie u1 true_atom
          else kk_nat_less (to_integer u1) (to_integer u2)
        else
          kk_int_less (to_integer u1) (to_integer u2)
      | Op2 (Subset, _, Opt (Atom 2), u1, u2) =>
        let
          val min_R = min_rep (rep_of u1) (rep_of u2)
          val r1 = to_rep min_R u1
          val r2 = to_rep min_R u2
        in
          case min_R of
            Opt (Vect (k, Atom 2)) =>
            let
              val r = to_guard [u1, u2] (Atom 2)
                               (formula_to_atom kk (kk_vect_set_bool_op kk
                                                        kk_implies k r1 r2))
            in
              if is_opt_rep (rep_of u1) then kk_intersect r false_atom else r
            end
          | Func (Unit, Opt (Atom 2)) =>
            (is_opt_rep (rep_of u1) ? kk_intersect false_atom)
                (kk_implies3 kk r1 r2)
          | Func (R1, Opt (Atom 2)) =>
            double_rel_let kk
                (fn r1 => fn r2 =>
                    kk_rel_if
                        (kk_some
                             (kk_intersect
                                  (kk_join r1 true_atom)
                                  (kk_join r2 false_atom)))
                        false_atom
                        (if is_opt_rep (rep_of u1) then
                           Kodkod.None
                         else
                           kk_rel_if
                               (kk_rel_eq (kk_union (kk_join r1 false_atom)
                                                    (kk_join r2 true_atom))
                                          (full_rel_for_rep R1))
                               true_atom Kodkod.None))
                r1 r2
          | _ => raise BadRep ("to_set_bool3_op", [min_R])
        end
      | Op2 (Eq, T, Opt (Atom 2), u1, u2) =>
        if is_Biggie u1 then
          to_compare_with_biggie u2 false_atom
        else if is_Biggie u2 then
          to_compare_with_biggie u1 false_atom
        else
          let
            val min_R = min_rep (rep_of u1) (rep_of u2)
            val r1 = to_rep min_R u1
            val r2 = to_rep min_R u2
          in
            if arity_of_rep min_R = 1 then
              kk_eq3 r1 r2
            else
              to_guard [u1, u2] (Atom 2) (formula_to_atom kk (kk_rel_eq r1 r2))
          end
      | Op2 (Union, _, R, u1, u2) =>
        to_set_op kk_or (kk_or3 kk) kk_union kk_intersect false R u1 u2
      | Op2 (SetDifference, _, R, u1, u2) =>
        to_set_op (kk_notimplies kk) (kk_notimplies3 kk) kk_difference
                  (fn r1 => fn r2 => kk_difference r2 r1) true R u1 u2
      | Op2 (Intersect, _, R, u1, u2) =>
        to_set_op kk_and (kk_and3 kk) kk_intersect kk_union false R u1 u2
      | Op2 (Apply, @{typ nat}, _,
             Op2 (Apply, _, _, Cst (Subtract, _, _), u1), u2) =>
        if is_Biggie u2 andalso not (is_opt_rep (rep_of u1)) then atom_0
        else fold kk_join [to_integer u1, to_integer u2] nat_subtract_rel
      | Op2 (Apply, _, R, u1, u2) =>
        if is_Biggie u2 andalso is_set_type (type_of u1)
           andalso not (is_opt_rep (rep_of u1)) then
          false_atom
        else
          to_apply R u1 u2
      | Op2 (Abstraction, _, Opt (Atom 1), u1, u2) =>
        to_guard [u1, u2] (Atom 1) atom_0
      | Op2 (Abstraction, T, Func (_, Formula), u1, u2) =>
        kk_comprehension (untuple to_decl u1) (to_f u2)
      | Op2 (Abstraction, T, Func (_, R2), u1, u2) =>
        let
          val dom_decls = untuple to_decl u1
          val ran_schema = atom_schema_of_rep R2
          val ran_decls = decls_for_atom_schema ~1 ran_schema
          val ran_vars = unary_var_seq ~1 (length ran_decls)
        in
          kk_comprehension (dom_decls @ ran_decls)
                           (kk_subset (fold1 kk_product ran_vars)
                                      (to_rep R2 u2))
        end
      | Op3 (Let, _, R, u1, u2, u3) =>
        kk_rel_let [to_expr_assign u1 u2] (to_rep R u3)
      | Op3 (If, _, R, u1, u2, u3) =>
        if rep_of u1 = Opt (Atom 2) then
          triple_rel_let kk
              (fn r1 => fn r2 => fn r3 =>
                  kk_rel_if (kk_some r1)
                            (kk_rel_if (atom_to_formula kk r1) r2 r3)
                            (kk_rel_if (kk_rel_eq r2 r3) r2
                                       (empty_rel_for_rep R)))
              (to_r u1) (to_rep R u2) (to_rep R u3)
        else
          kk_rel_if (to_f u1) (to_rep R u2) (to_rep R u3)
      | Tuple (_, Struct Rs, us) => to_product Rs us
      | Tuple (_, Vect (k, R), us) => to_product (map rep_of us) us
      | Tuple (_, Opt R, us) =>
        let
          val Rs = map rep_of us
          val Rs' = map unopt_rep Rs
        in rel_expr_to_rep kk R (Struct Rs') (to_product Rs us) end
      | BoundRel (x, _, _, _) => Kodkod.Var x
      | FreeRel (x, _, _, _) => Kodkod.Rel x
      | RelReg (j, _, R) => Kodkod.RelReg (arity_of_rep R, j)
      | FormulaReg (j, T, Atom 2) => to_atom 2 (FormulaReg (j, T, Formula))
      | u => raise BadTerm ("to_r", [u])
    (* term -> Kodkod.decl *)
    and to_decl (BoundRel (x, _, R, _)) =
        Kodkod.DeclOne (x, Kodkod.AtomSeq (card_of_rep R, 0))
      | to_decl u = raise BadTerm ("to_decl", [u])
    (* term -> Kodkod.expr_assign *)
    and to_expr_assign (FormulaReg (j, _, R)) u =
        Kodkod.AssignFormulaReg (j, to_f u)
      | to_expr_assign (RelReg (j, _, R)) u =
        Kodkod.AssignRelReg ((arity_of_rep R, j), to_r u)
      | to_expr_assign u1 _ = raise BadTerm ("to_expr_assign", [u1])
    (* int -> term -> Kodkod.rel_expr *)
    and to_atom k u =
      case rep_of u of
        Formula => formula_to_atom kk (to_f u)
      | Unit => atom_0
      | R => rel_expr_to_atom kk k R (to_r u)
    (* rep list -> term -> Kodkod.rel_expr *)
    and to_struct Rs u = rel_expr_to_struct kk Rs (rep_of u) (to_r u)
    (* int -> rep -> term -> Kodkod.rel_expr *)
    and to_vect k R u = rel_expr_to_vect kk k R (rep_of u) (to_r u)
    (* rep -> rep -> term -> Kodkod.rel_expr *)
    and to_func R1 R2 u = rel_expr_to_func kk R1 R2 (rep_of u) (to_r u)
    (* rep -> term -> Kodkod.rel_expr *)
    and to_opt R u =
      let val old_R = rep_of u in
        if is_opt_rep old_R then rel_expr_to_rep kk (Opt R) old_R (to_r u)
        else to_rep R u
      end
    (* rep -> term -> Kodkod.rel_expr *)
    and to_rep (Atom k) u = to_atom k u
      | to_rep (Struct Rs) u = to_struct Rs u
      | to_rep (Vect (k, R)) u = to_vect k R u
      | to_rep (Func (R1, R2)) u = to_func R1 R2 u
      | to_rep (Opt R) u = to_opt R u
      | to_rep R _ = raise BadRep ("to_rep", [R])
    (* term -> Kodkod.rel_expr *)
    and to_integer u =
      let
        val R = rep_of u
        val R' = (is_opt_rep R ? Opt) (Atom (card_of_rep R))
      in to_rep R' u end
    (* term list -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
    and to_guard guard_us R r =
      let
        val unpacked_rs = unpack_joins r
        val plain_guard_rs =
          map to_r (filter (is_Opt o rep_of) guard_us)
          |> filter (not o member op= unpacked_rs)
        val func_guard_us =
          filter ((is_Func andf is_opt_rep) o rep_of) guard_us
        val func_guard_rs = map to_r func_guard_us
        val guard_fs =
            map kk_no plain_guard_rs @
            map2 (kk_not oo kk_n_ary_function peephole_optim kk)
                 (map (unopt_rep o rep_of) func_guard_us) func_guard_rs
      in
        if null guard_fs then
          r
        else
          kk_rel_if (fold1 kk_or guard_fs) (empty_rel_for_rep R) r
      end
    (* rep -> rep -> Kodkod.rel_expr -> int -> Kodkod.rel_expr *)
    and to_project new_R old_R r j0 =
      rel_expr_to_rep kk new_R old_R (kk_project_seq r j0 (arity_of_rep old_R))
    (* rep list -> term list -> Kodkod.rel_expr *)
    and to_product Rs us =
      case map (uncurry to_rep) (filter (not_equal Unit o fst) (Rs ~~ us)) of
        [] => raise BadRep ("to_product", Rs)
      | rs => fold1 kk_product rs
    (* (Kodkod.formula -> Kodkod.formula -> Kodkod.formula)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula) -> term -> term
       -> Kodkod.rel_expr *)
    and to_set_bool_op connective set_oper u1 u2 =
      let
        val min_R = min_rep (rep_of u1) (rep_of u2)
        val r1 = to_rep min_R u1
        val r2 = to_rep min_R u2
      in
        case min_R of
          Vect (k, Atom 2) => kk_vect_set_bool_op kk connective k r1 r2
        | Func (R1, Formula) => set_oper r1 r2
        | Func (R1, Atom 2) => set_oper (kk_join r1 true_atom)
                                        (kk_join r2 true_atom)
        | _ => raise BadRep ("to_set_bool_op", [min_R])
      end
    (* (Kodkod.formula -> Kodkod.formula -> Kodkod.formula)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula) -> bool -> rep
       -> term -> term -> Kodkod.rel_expr *)
    and to_set_op connective connective3 set_oper comp_set_oper neg_second R
                  u1 u2 =
      let
        val min_R = min_rep (rep_of u1) (rep_of u2)
        val r1 = to_rep min_R u1
        val r2 = to_rep min_R u2
        val unopt_R = unopt_rep R
      in
        to_guard [u1, u2] unopt_R
            (rel_expr_to_rep kk unopt_R (unopt_rep min_R)
                 (case min_R of
                   Opt (Vect (k, Atom 2)) =>
                   kk_vect_set_op kk connective k r1 r2
                 | Vect (k, Atom 2) => kk_vect_set_op kk connective k r1 r2
                 | Func (_, Formula) => set_oper r1 r2
                 | Func (Unit, Opt (Atom 2)) => connective3 r1 r2
                 | Func (R1, Opt (Atom 2)) =>
                   double_rel_let kk
                       (fn r1 => fn r2 =>
                           kk_union
                               (kk_product
                                    (set_oper (kk_join r1 true_atom)
                                              (kk_join r2 (bool_to_atom
                                                             (not neg_second))))
                                    true_atom)
                               (kk_product
                                    (comp_set_oper (kk_join r1 false_atom)
                                                   (kk_join r2 (bool_to_atom
                                                                   neg_second)))
                                    false_atom))
                       r1 r2
                 | _ => raise BadRep ("to_set_op", [min_R])))
      end
    (* rep -> rep -> Kodkod.rel_expr -> term -> Kodkod.rel_expr *)
    and to_apply res_R func_u arg_u =
      case unopt_rep (rep_of func_u) of
        Unit =>
        to_guard [arg_u] res_R (rel_expr_to_rep kk res_R (Atom 1) atom_0)
      | Atom 1 =>
        to_guard [arg_u] res_R (rel_expr_to_rep kk res_R (Atom 1) (to_r func_u))
      | Atom k =>
        let
          val dom_card = card_of_rep (rep_of arg_u)
          val ran_R = Atom (exact_root dom_card k)
        in
          to_apply_vect dom_card ran_R res_R (to_vect dom_card ran_R func_u)
                        arg_u
        end
      | Vect (1, R') =>
        to_guard [arg_u] res_R (rel_expr_to_rep kk res_R R' (to_r func_u))
      | Vect (k, R') => to_apply_vect k R' res_R (to_r func_u) arg_u
      | Func (R, Formula) =>
        to_guard [arg_u] res_R (formula_to_rel_expr kk res_R
                                    (kk_subset (to_opt R arg_u) (to_r func_u)))
      | Func (Unit, R2) =>
        to_guard [arg_u] res_R (rel_expr_to_rep kk res_R R2 (to_r func_u))
      | Func (R1, R2) =>
        rel_expr_to_rep kk res_R R2
            (kk_n_fold_join kk (arity_of_rep R1) R2 (to_opt R1 arg_u)
                               (to_r func_u))
      | R => raise BadRep ("to_apply", [R])
    (* int -> rep -> rep -> Kodkod.rel_expr -> term *)
    and to_apply_vect k R' res_R func_r arg_u =
      let
        val arg_R = unopt_rep (rep_of arg_u)
        val arg_R = if is_one_rep arg_R then arg_R
                    else Atom (card_of_rep arg_R)
        val vect_r = rel_expr_to_vect kk k res_R (Vect (k, R')) func_r
        val vect_rs = unpack_vect_in_chunks kk (arity_of_rep res_R) k vect_r
      in
        kk_case_switch kk (to_opt arg_R arg_u) (all_singletons_for_rep arg_R)
                       vect_rs
      end
    (* term -> Kodkod.rel_expr -> Kodkod.rel_expr *)
    and to_compare_with_biggie u r =
      if is_opt_rep (rep_of u) then kk_join (to_r u) (kk_product Kodkod.Univ r)
      else r
  in
    case rep_of u of
      Opt (Atom 2) => if use_result_rel then
                        kk_rel_eq ((not negate ? kk_not3) (to_r u)) result_rel
                      else if liberal then
                        kk_not (kk_rel_eq (to_r u) (bool_to_atom negate))
                      else
                        kk_rel_eq (to_r u) (bool_to_atom (not negate))
    |  _ => (negate ? kk_not) (to_f u)
  end

end;
