(*  Title:      Nitpick/Tools/nitpick_term.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Intermediate Nitpick terms.
*)

signature NITPICK_TERM =
sig
  type hol_term = Term.term
  type scope = NitpickScope.scope
  type name_pool = NitpickPeephole.name_pool
  type rep = NitpickRep.rep

  datatype cst =
    Unity |
    False |
    True |
    None |
    Univ |
    Iden |
    Num of int |
    Biggie |
    Suc |
    Add |
    Subtract |
    Multiply |
    NatToInt |
    IntToNat

  datatype op1 =
    Not |
    Finite |
    Card |
    Closure |
    SingletonSet |
    First |
    Second |
    Cast

  datatype op2 =
    All |
    Exist |
    Or |
    Implies |
    And |
    Less |
    Subset |
    StrictEq |
    Eq |
    Union |
    SetDifference |
    Intersect |
    Apply |
    Abstraction

  datatype op3 =
    Let |
    If

  datatype term =
    Cst of cst * typ * rep |
    Op1 of op1 * typ * rep * term |
    Op2 of op2 * typ * rep * term * term |
    Op3 of op3 * typ * rep * term * term * term |
    Tuple of typ * rep * term list |
    BoundName of int * typ * rep * string |
    FreeName of string * typ * rep |
    ConstName of string * typ * rep |
    BoundRel of Kodkod.n_ary_index * typ * rep * string |
    FreeRel of Kodkod.n_ary_index * typ * rep * string |
    RelReg of int * typ * rep |
    FormulaReg of int * typ * rep

  structure NameTable : TABLE

  exception BadTerm of string * term list

  val string_for_term : Proof.context -> term -> string
  val inline_term : term -> bool
  val type_of : term -> typ
  val rep_of : term -> rep
  val nickname_of : term -> string
  val is_ctor_name : theory -> term -> bool
  val is_skolem_name : term -> bool
  val is_Biggie : term -> bool
  val fold_term : (term -> 'a -> 'a) -> term -> 'a -> 'a
  val untuple : (term -> 'a) -> term -> 'a list
  val add_free_names : term -> term list -> term list
  val add_const_names : term -> term list -> term list
  val name_ord : (term * term) -> order
  val lookup_name : 'a NameTable.table -> term -> 'a
  val lookup_rel : term NameTable.table -> term -> Kodkod.n_ary_index
  val from_hol_term : theory -> op2 -> hol_term -> term
  val choose_reps_for_free_vars :
    scope -> term list -> rep NameTable.table -> term list * rep NameTable.table
  val choose_reps_for_consts :
    scope -> bool -> term list -> rep NameTable.table
    -> term list * rep NameTable.table
  val choose_reps_for_all_ctors :
    scope -> rep NameTable.table -> term list * rep NameTable.table
  val choose_reps_in_term :
    bool -> rep -> scope -> rep NameTable.table -> term -> term
  val rename_free_vars :
    term list -> name_pool -> term NameTable.table
    -> term list * name_pool * term NameTable.table
  val rename_vars_in_term : name_pool -> term NameTable.table -> term -> term
end;

structure NitpickTerm : NITPICK_TERM =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep

datatype cst =
  Unity |
  False |
  True |
  None |
  Univ |
  Iden |
  Num of int |
  Biggie |
  Suc |
  Add |
  Subtract |
  Multiply |
  NatToInt |
  IntToNat

datatype op1 =
  Not |
  Finite |
  Card |
  Closure |
  SingletonSet |
  First |
  Second |
  Cast

datatype op2 =
  All |
  Exist |
  Or |
  Implies |
  And |
  Less |
  Subset |
  StrictEq |
  Eq |
  Union |
  SetDifference |
  Intersect |
  Apply |
  Abstraction

datatype op3 =
  Let |
  If

datatype term =
  Cst of cst * typ * rep |
  Op1 of op1 * typ * rep * term |
  Op2 of op2 * typ * rep * term * term |
  Op3 of op3 * typ * rep * term * term * term |
  Tuple of typ * rep * term list |
  BoundName of int * typ * rep * string |
  FreeName of string * typ * rep |
  ConstName of string * typ * rep |
  BoundRel of Kodkod.n_ary_index * typ * rep * string |
  FreeRel of Kodkod.n_ary_index * typ * rep * string |
  RelReg of int * typ * rep |
  FormulaReg of int * typ * rep

exception BadTerm of string * term list

(* cst -> string *)
fun string_for_cst Unity = "Unity"
  | string_for_cst False = "False"
  | string_for_cst True = "True"
  | string_for_cst None = "None"
  | string_for_cst Univ = "Univ"
  | string_for_cst Iden = "Iden"
  | string_for_cst (Num j) = "Num " ^ Int.toString j
  | string_for_cst Biggie = "Biggie"
  | string_for_cst Suc = "Suc"
  | string_for_cst Add = "Add"
  | string_for_cst Subtract = "Subtract"
  | string_for_cst Multiply = "Multiply"
  | string_for_cst NatToInt = "NatToInt"
  | string_for_cst IntToNat = "IntToNat"

(* op1 -> string *)
fun string_for_op1 Not = "Not"
  | string_for_op1 Finite = "Finite"
  | string_for_op1 Card = "Card"
  | string_for_op1 Closure = "Closure"
  | string_for_op1 SingletonSet = "SingletonSet"
  | string_for_op1 First = "First"
  | string_for_op1 Second = "Second"
  | string_for_op1 Cast = "Cast"

(* op2 -> string *)
fun string_for_op2 All = "All"
  | string_for_op2 Exist = "Exist"
  | string_for_op2 Or = "Or"
  | string_for_op2 Implies = "Implies"
  | string_for_op2 And = "And"
  | string_for_op2 Less = "Less"
  | string_for_op2 Subset = "Subset"
  | string_for_op2 StrictEq = "StrictEq"
  | string_for_op2 Eq = "Eq"
  | string_for_op2 Union = "Union"
  | string_for_op2 SetDifference = "SetDifference"
  | string_for_op2 Intersect = "Intersect"
  | string_for_op2 Apply = "Apply"
  | string_for_op2 Abstraction = "Abstraction"

(* op3 -> string *)
fun string_for_op3 Let = "Let"
  | string_for_op3 If = "If"

(* int -> Proof.context -> term -> string *)
fun basic_string_for_term indent ctxt u =
  let
    (* term -> string *)
    val sub = basic_string_for_term (indent + 1) ctxt
  in
    (if indent = 0 then "" else "\n" ^ StringCvt.padLeft #" " (2 * indent) "") ^
    "(" ^
    (case u of
       Cst (c, T, R) =>
       "Cst " ^ string_for_cst c ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R
     | Op1 (oper, T, R, u1) =>
       "Op1 " ^ string_for_op1 oper ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ sub u1
     | Op2 (oper, T, R, u1, u2) =>
       "Op2 " ^ string_for_op2 oper ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ sub u1 ^ " " ^ sub u2
     | Op3 (oper, T, R, u1, u2, u3) =>
       "Op3 " ^ string_for_op3 oper ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ sub u1 ^ " " ^ sub u2 ^ " " ^ sub u3
     | Tuple (T, R, us) =>
       "Tuple " ^ Syntax.string_of_typ ctxt T ^ " " ^ string_for_rep R ^
       foldr op^ "" (map sub us)
     | BoundName (j, T, R, nick) =>
       "BoundName " ^ Int.toString j ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ nick
     | FreeName (s, T, R) =>
       "FreeName " ^ s ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R
     | ConstName (s, T, R) =>
       "ConstName " ^ s ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R
     | BoundRel ((n, j), T, R, nick) =>
       "BoundRel " ^ Int.toString n ^ "." ^ Int.toString j ^ " " ^
       Syntax.string_of_typ ctxt T ^ " " ^ string_for_rep R ^ " " ^ nick
     | FreeRel ((n, j), T, R, nick) =>
       "FreeRel " ^ Int.toString n ^ "." ^ Int.toString j ^ " " ^
       Syntax.string_of_typ ctxt T ^ " " ^ string_for_rep R ^ " " ^ nick
     | RelReg (j, T, R) =>
       "RelReg " ^ Int.toString j ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R
     | FormulaReg (j, T, R) =>
       "FormulaReg " ^ Int.toString j ^ " " ^ Syntax.string_of_typ ctxt T ^
       " " ^ string_for_rep R) ^
    ")"
  end
(* Proof.context -> term -> string *)
val string_for_term = basic_string_for_term 0

(* term -> bool *)
fun inline_term (Op1 _) = false
  | inline_term (Op2 _) = false
  | inline_term (Op3 _) = false
  | inline_term (Tuple (_, _, us)) = forall inline_term us
  | inline_term _ = true

(* term -> typ *)
fun type_of (Cst (_, T, _)) = T
  | type_of (Op1 (_, T, _, _)) = T
  | type_of (Op2 (_, T, _, _, _)) = T
  | type_of (Op3 (_, T, _, _, _, _)) = T
  | type_of (Tuple (T, _, _)) = T
  | type_of (BoundName (_, T, _, _)) = T
  | type_of (FreeName (_, T, _)) = T
  | type_of (ConstName (_, T, _)) = T
  | type_of (BoundRel (_, T, _, _)) = T
  | type_of (FreeRel (_, T, _, _)) = T
  | type_of (RelReg (_, T, _)) = T
  | type_of (FormulaReg (_, T, _)) = T

(* term -> rep *)
fun rep_of (Cst (_, _, R)) = R
  | rep_of (Op1 (_, _, R, _)) = R
  | rep_of (Op2 (_, _, R, _, _)) = R
  | rep_of (Op3 (_, _, R, _, _, _)) = R
  | rep_of (Tuple (_, R, _)) = R
  | rep_of (BoundName (_, _, R, _)) = R
  | rep_of (FreeName (_, _, R)) = R
  | rep_of (ConstName (_, _, R)) = R
  | rep_of (BoundRel (_, _, R, _)) = R
  | rep_of (FreeRel (_, _, R, _)) = R
  | rep_of (RelReg (_, _, R)) = R
  | rep_of (FormulaReg (_, _, R)) = R

(* term -> string *)
fun nickname_of (BoundName (_, _, _, nick)) = nick
  | nickname_of (FreeName (s, _, _)) = s
  | nickname_of (ConstName (s, _, _)) = s
  | nickname_of u = raise BadTerm ("nickname_of", [u])

(* theory -> term -> bool *)
fun is_ctor_name thy (ConstName (s, T, _)) = is_ctor thy (s, T)
  | is_ctor_name _ _ = false
(* term -> bool *)
fun is_skolem_name (ConstName (s, _, _)) = String.isPrefix skolem_prefix s
  | is_skolem_name _ = false
(* term -> bool *)
fun is_Biggie (Cst (Biggie, _, _)) = true
  | is_Biggie _ = false
(* theory -> term -> bool *)
fun is_biggie thy (Op2 (Apply, _, _, u1, u2)) =
    is_biggie thy u2 andalso is_ctor_name thy u1
  | is_biggie _ u = is_Biggie u

(* (term -> 'a -> 'a) -> term -> 'a -> 'a *)
fun fold_term f u =
  case u of
    Op1 (_, _, _, u1) => fold_term f u1
  | Op2 (_, _, _, u1, u2) => fold_term f u1 #> fold_term f u2
  | Op3 (_, _, _, u1, u2, u3) =>
    fold_term f u1 #> fold_term f u2 #> fold_term f u3
  | Tuple (_, _, us) => fold (fold_term f) us
  | _ => f u

(* term -> term -> int *)
fun num_occurrences_in_term needle_u stack_u =
  fold_term (fn u => if u = needle_u then curry op+ 1 else I) stack_u 0
(* term -> term -> bool *)
val is_subterm_of = not_equal 0 oo num_occurrences_in_term

(* term -> term *)
fun optimize_unit u =
  if rep_of u = Unit then Cst (Unity, type_of u, Unit) else u
(* theory -> term -> term *)
fun optimize_biggie thy u =
  if is_biggie thy u then Cst (Biggie, type_of u, rep_of u) else u

(* theory -> term -> term -> term -> term *)
fun substitute_in_term thy needle_u needle_u' =
  let
    (* term -> term *)
    fun aux stack_u =
      if stack_u = needle_u then
        needle_u'
      else
        case stack_u of
          Op1 (oper, T, R, u1) => Op1 (oper, T, R, aux u1)
        | Op2 (oper, T, R, u1, u2) =>
          optimize_biggie thy (Op2 (oper, T, R, aux u1, aux u2))
        | Op3 (oper, T, R, u1, u2, u3) =>
          Op3 (oper, T, R, aux u1, aux u2, aux u3)
        | Tuple (T, R, us) => Tuple (T, R, map aux us)
        |  _ => stack_u
  in aux end

(* theory -> term -> term -> term -> term *)
fun optimized_let thy u1 u2 u3 =
  if inline_term u2 orelse num_occurrences_in_term u1 u3 < 2 then
    substitute_in_term thy u1 u2 u3
  else
    Op3 (Let, type_of u3, rep_of u3, u1, u2, u3)

(* (term -> 'a) -> term -> 'a list *)
fun untuple f (Tuple (_, _, us)) = List.concat (map (untuple f) us)
  | untuple f u = if rep_of u = Unit then [] else [f u]

(* term -> term list -> term list *)
val add_free_names =
  fold_term (fn u => fn vars =>
                case u of FreeName _ => insert op= u vars | _ => vars)
val add_const_names =
  fold_term (fn u => fn vars =>
                case u of ConstName _ => insert op= u vars | _ => vars)

(* term -> rep -> term *)
fun change_name_rep (BoundName (j, T, _, nick)) R = BoundName (j, T, R, nick)
  | change_name_rep (FreeName (s, T, _)) R = FreeName (s, T, R)
  | change_name_rep (ConstName (s, T, _)) R = ConstName (s, T, R)
  | change_name_rep u _ = raise BadTerm ("change_name_rep", [u])

(* (term * term) -> order *)
fun name_ord (BoundName (j1, _, _, _), BoundName (j2, _, _, _)) =
    int_ord (j1, j2)
  | name_ord (BoundName _, _) = LESS
  | name_ord (_, BoundName _) = GREATER
  | name_ord (FreeName (s1, _, _), FreeName (s2, _, _)) =
    fast_string_ord (s1, s2)
  | name_ord (FreeName _, _) = LESS
  | name_ord (_, FreeName _) = GREATER
  | name_ord (ConstName (s1, T1, _), ConstName (s2, T2, _)) =
    (case fast_string_ord (s1, s2) of
       EQUAL => TermOrd.typ_ord (T1, T2)
     | ord => ord)
  | name_ord (u1, u2) = raise BadTerm ("name_ord", [u1, u2])

structure NameTable = TableFun(type key = term val ord = name_ord)

(* 'a NameTable.table -> term -> 'a *)
fun lookup_name table name =
  case NameTable.lookup table name of
    SOME u => u
  | NONE => raise BadTerm ("lookup_name", [name])
(* term NameTable.table -> term -> Kodkod.n_ary_index *)
fun lookup_rel table name =
  case lookup_name table name of
    FreeRel (x, _, _, _) => x
  | u => raise BadTerm ("lookup_rel", [u])

(* theory -> op2 -> hol_term -> term *)
fun from_hol_term thy eq =
  let
    (* string list -> typ list -> hol_term -> term *)
    fun aux eq ss Ts t =
      let
        (* hol_term -> term *)
        val sub = aux eq ss Ts
        (* string -> typ -> hol_term -> term *)
        fun sub_abs s T = aux eq (s :: ss) (T :: Ts)
        (* typ -> hol_term -> hol_term -> term *)
        fun sub_equals T t1 t2 =
          let
            val (binder_Ts, body_T) = strip_type (domain_type T)
            val n = length binder_Ts
          in
            if eq = Eq andalso n > 0 then
              let
                val t1 = incr_boundvars n t1
                val t2 = incr_boundvars n t2
                val xs = map Bound (n - 1 downto 0)
                val equation = Const (@{const_name "op ="},
                                      body_T --> body_T --> bool_T)
                                   $ betapplys (t1, xs) $ betapplys (t2, xs)
                val (t, _) =
                  foldr (fn (T, (t, j)) =>
                            (Const (@{const_name All}, T --> bool_T)
                             $ Abs ("x" ^ nat_subscript j, T, t),
                             j - 1))
                        (equation, n) binder_Ts
              in sub t end
            else
              Op2 (eq, bool_T, Any, aux Eq ss Ts t1, aux Eq ss Ts t2)
          end
        (* op2 -> string -> typ -> hol_term -> term *)
        fun handle_quantifier quant s T t1 =
          let
            val bound_u = BoundName (length Ts, T, Any, s)
            val body_u = sub_abs s T t1
          in
            if is_subterm_of bound_u body_u then
              Op2 (quant, bool_T, Any, bound_u, body_u)
            else
              body_u
          end
      in
        case t of
        (* Pure *)
          Const (@{const_name all}, _) $ Abs (s, T, t1) =>
          handle_quantifier All s T t1
        | (t0 as Const (@{const_name all}, T)) $ t1 =>
          sub (t0 $ eta_expand Ts t1 1)
        | Const (@{const_name "=="}, T) $ t1 $ t2 => sub_equals T t1 t2
        | Const (@{const_name "==>"}, _) $ t1 $ t2 =>
          Op2 (Implies, bool_T, Any, sub t1, sub t2)
        (* HOL *)
        | Const (@{const_name Trueprop}, _) $ t1 => sub t1
        | Const (@{const_name Not}, _) $ t1 =>
          (case sub t1 of
             Op1 (Not, _, _, u11) => u11
           | u1 => Op1 (Not, bool_T, Any, u1))
        | Const (@{const_name False}, T) => Cst (False, T, Any)
        | Const (@{const_name True}, T) => Cst (True, T, Any)
        | Const (@{const_name All}, _) $ Abs (s, T, t1) =>
          handle_quantifier All s T t1
        | (t0 as Const (@{const_name All}, T)) $ t1 =>
          sub (t0 $ eta_expand Ts t1 1)
        | Const (@{const_name Ex}, _) $ Abs (s, T, t1) =>
          handle_quantifier Exist s T t1
        | (t0 as Const (@{const_name Ex}, T)) $ t1 =>
          sub (t0 $ eta_expand Ts t1 1)
        | Const (@{const_name "op ="}, T) $ t1 $ t2 => sub_equals T t1 t2
        | Const (@{const_name "op &"}, _) $ t1 $ t2 =>
          Op2 (And, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name "op |"}, _) $ t1 $ t2 =>
          Op2 (Or, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name "op -->"}, _) $ t1 $ t2 =>
          Op2 (Implies, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name If}, T) $ t1 $ t2 $ t3 =>
          Op3 (If, third_range_type T, Any, sub t1, sub t2, sub t3)
        | Const (@{const_name Let}, T) $ t1 $ Abs (s, T', t2) =>
          Op3 (Let, second_range_type T, Any, BoundName (length Ts, T', Any, s),
               sub t1, sub_abs s T' t2)
        | (t0 as Const (@{const_name Let}, T)) $ t1 $ t2 =>
          sub (t0 $ t1 $ eta_expand Ts t2 1)
        | @{const Unity} => Cst (Unity, @{typ unit}, Any)
        | Const (@{const_name Pair}, T) $ t1 $ t2 =>
          Tuple (second_range_type T, Any, map (sub) [t1, t2])
        | Const (@{const_name fst}, T) $ t1 =>
          Op1 (First, range_type T, Any, sub t1)
        | Const (@{const_name snd}, T) $ t1 =>
          Op1 (Second, range_type T, Any, sub t1)
        | Const (@{const_name Id}, T) => Cst (Iden, T, Any)
        | Const (@{const_name Set.empty}, T) => Cst (None, T, Any)
        | Const (@{const_name Set.UNIV}, T) => Cst (Univ, T, Any)
        | Const (@{const_name "op :"}, _) $ t1 $ t2 => sub (t2 $ t1)
        | Const (@{const_name insert}, T) $ t1 $ t2 =>
          (case t2 of
             Const (@{const_name Set.empty}, _) =>
             Op1 (SingletonSet, second_range_type T, Any, sub t1)
           | _ =>
             Op2 (Union, second_range_type T, Any,
                  Op1 (SingletonSet, second_range_type T, Any, sub t1), sub t2))
        | Const (@{const_name trancl}, T) $ t1 =>
          Op1 (Closure, range_type T, Any, sub t1)
        | Const (@{const_name rtrancl}, T) $ t1 =>
          let val ran_T = range_type T in
            Op2 (Union, ran_T, Any, Op1 (Closure, ran_T, Any, sub t1),
                 Cst (Iden, ran_T, Any))
          end
        | Const (@{const_name Set.Un}, T) $ t1 $ t2 =>
          Op2 (Union, second_range_type T, Any, sub t1, sub t2)
        | Const (@{const_name Set.Int}, T) $ t1 $ t2 =>
          Op2 (Intersect, second_range_type T, Any, sub t1, sub t2)
        | Const (s as @{const_name zero_class.zero}, T) =>
          if is_integer_type T then Cst (Num 0, T, Any)
          else ConstName (s, T, Any)
        | Const (s as @{const_name one_class.one}, T) =>
          if is_integer_type T then Cst (Num 1, T, Any)
          else ConstName (s, T, Any)
        | Const (@{const_name Suc}, T) => Cst (Suc, T, Any)
        | Const (s as @{const_name plus_class.plus}, T) =>
          if is_integer_type (domain_type T) then Cst (Add, T, Any)
          else ConstName (s, T, Any)
        | (t0 as Const (@{const_name minus_class.minus}, T)) $ t1 $ t2 =>
          if is_set_type (domain_type T) then
            Op2 (SetDifference, second_range_type T, Any, sub t1, sub t2)
          else
            Op2 (Apply, second_range_type T, Any, sub (t0 $ t1), sub t2)
        | Const (s as @{const_name minus_class.minus}, T) =>
          if is_integer_type (domain_type T) then Cst (Subtract, T, Any)
          else ConstName (s, T, Any)
        | Const (s as @{const_name times_class.times}, T) =>
          if is_integer_type (domain_type T) then Cst (Multiply, T, Any)
          else ConstName (s, T, Any)
        | (t0 as Const (@{const_name ord_class.less}, T)) $ t1 $ t2 =>
          let val dom_T = domain_type T in
            if is_set_type dom_T then
              let
                val j = length Ts
                val v1 = BoundName (j, dom_T, Any, "less.x1")
                val v2 = BoundName (j + 1, dom_T, Any, "less.x2")
              in
                Op3 (Let, bool_T, Any, v1, sub t1,
                     Op3 (Let, bool_T, Any, v2, sub t2,
                          Op2 (And, bool_T, Any,
                               Op2 (Subset, bool_T, Any, v1, v2),
                               Op1 (Not, bool_T, Any,
                                    Op2 (Subset, bool_T, Any, v2, v1)))))
              end
            else if is_integer_type dom_T then
              let val u2 = sub t2 in
                (case u2 of
                   Cst (Num j, _, _) =>
                   if j > 0 then
                     Op1 (Not, bool_T, Any,
                          Op2 (Less, bool_T, Any, Cst (Num (j - 1), dom_T, Any),
                               sub t1))
                   else
                     raise Same
                 | _ => raise Same)
                handle Same => Op2 (Less, bool_T, Any, sub t1, u2)
              end
            else
              Op2 (Apply, second_range_type T, Any, sub (t0 $ t1), sub t2)
          end
        | (t0 as Const (@{const_name ord_class.less_eq}, T)) $ t1 $ t2 =>
          let val dom_T = domain_type T in
            if is_set_type dom_T then
              Op2 (Subset, bool_T, Any, sub t1, sub t2)
            else if is_integer_type dom_T then
              Op1 (Not, bool_T, Any, Op2 (Less, bool_T, Any, sub t2, sub t1))
            else
              Op2 (Apply, second_range_type T, Any, sub (t0 $ t1), sub t2)
          end
        | Const (@{const_name uminus_class.uminus},
                 Type ("fun", [@{typ int}, @{typ int}])) =>
          Op2 (Apply, int_T --> int_T, Any,
               Cst (Subtract, [int_T, int_T] ---> int_T, Any),
               Cst (Num 0, int_T, Any))
        | Const (@{const_name uminus_class.uminus},
                 Type ("fun", [T1 as Type ("fun", [_, @{typ bool}]), _]))
          $ t1 =>
          Op2 (SetDifference, T1, Any, Cst (Univ, T1, Any), sub t1)
        | Const (@{const_name of_nat}, T as @{typ "nat => int"}) =>
          Cst (NatToInt, T, Any)
        | Const (@{const_name nat}, T) => Cst (IntToNat, T, Any)
        | Const (@{const_name finite}, _) $ t1 =>
          Op1 (Finite, bool_T, Any, sub t1)
        | Const (@{const_name card}, _) $ t1 => Op1 (Card, nat_T, Any, sub t1)
        | Const (s, T) =>
          if String.isPrefix numeral_prefix s then
            Cst (Num (the (Int.fromString (unprefix numeral_prefix s))), T, Any)
          else
            ConstName (s, T, Any)
        | Free (s, T) => FreeName (s, T, Any)
        | Var _ => raise BadHOLTerm ("from_hol_term", [t])
        | Bound j => BoundName (length Ts - j - 1, nth Ts j, Any, nth ss j)
        | Abs (s, T, t1) =>
          Op2 (Abstraction, T --> fastype_of1 (T :: Ts, t1), Any,
               BoundName (length Ts, T, Any, s), sub_abs s T t1)
        | t1 $ t2 =>
          Op2 (Apply, range_type (fastype_of1 (Ts, t1)), Any, sub t1, sub t2)
      end
  in aux eq [] [] end

(* scope -> term -> term list * rep NameTable.table
   -> term list * rep NameTable.table *)
fun choose_rep_for_free_var scope v (vs, table) =
  let
    val R = best_non_opt_set_rep_for_type scope (type_of v)
    val v = change_name_rep v R
  in (v :: vs, NameTable.update (v, R) table) end
(* scope -> bool -> term -> term list * rep NameTable.table
   -> term list * rep NameTable.table *)
fun choose_rep_for_const scope all_precise v (vs, table) =
  let
    val R = (if all_precise orelse is_skolem_name v then
               best_non_opt_set_rep_for_type
             else
               best_opt_set_rep_for_type) scope (type_of v)
    val v = change_name_rep v R
  in (v :: vs, NameTable.update (v, R) table) end

(* scope -> term list -> rep NameTable.table
   -> term list * rep NameTable.table *)
fun choose_reps_for_free_vars scope vs table =
  fold (choose_rep_for_free_var scope) vs ([], table)
(* scope -> bool -> term list -> rep NameTable.table
   -> term list * rep NameTable.table *)
fun choose_reps_for_consts scope all_precise vs table =
  fold (choose_rep_for_const scope all_precise) vs ([], table)

(* scope -> ctor_spec -> term list * rep NameTable.table
   -> term list * rep NameTable.table *)
fun choose_rep_for_ctor scope {name, typ, total, ...} (vs, table) =
  let
    val R = (if total then best_non_opt_set_rep_for_type
             else best_opt_set_rep_for_type) scope typ
    val v = ConstName (name, typ, R)
  in (v :: vs, NameTable.update (v, R) table) end
(* scope -> dtype_spec -> term list * rep NameTable.table
   -> term list * rep NameTable.table *)
fun choose_rep_for_ctors_of_datatype scope {ctors, ...} (vs, table) =
  fold (choose_rep_for_ctor scope) (rev ctors) (vs, table)
(* scope -> rep NameTable.table -> term list * rep NameTable.table *)
fun choose_reps_for_all_ctors (scope as {datatypes, ...}) table =
  fold (choose_rep_for_ctors_of_datatype scope) (rev datatypes) ([], table)

(* scope -> term -> rep NameTable.table -> rep NameTable.table *)
fun choose_rep_for_bound_var scope v table =
  let val R = best_one_rep_for_type scope (type_of v) in
    NameTable.update (v, R) table
  end

(* theory -> term -> bool *)
fun is_constructive thy u =
  (not (is_fun_type (type_of u))
   andalso (not (is_opt_rep (rep_of u)) orelse is_Biggie u))
  orelse case u of
           Cst (Num _, _, _) => true
         | Cst (cst, _, _) => cst = Suc orelse cst = Add
         | Op2 (Apply, _, _, u1, u2) =>
           is_constructive thy u1 andalso is_constructive thy u2
         | Op3 (If, _, _, _, u2, u3) => 
           is_constructive thy u2 andalso is_constructive thy u3
         | ConstName _ => is_ctor_name thy u
         | _ => false

(* op2 -> term -> term list *)
fun phrases All (Op2 (Implies, _, _, u1, u2)) =
    phrases Exist u1 @ phrases All u2
  | phrases Exist (Op2 (And, _, _, u1, u2)) = maps (phrases Exist) [u1, u2]
  | phrases _ u = [u]

(* theory -> op2 -> term -> term -> bool *)
fun is_subterm_equation thy oper bound_u body_u =
  let
    (* term -> term -> bool *)
    fun good_eq u1 u2 =
      not (is_opt_rep (rep_of u1)) andalso is_constructive thy u2
      andalso is_subterm_of bound_u u2
  in
    case body_u of
      Op2 (StrictEq, _, _, u1, u2) =>
      rep_of bound_u <> Unit andalso (good_eq u1 u2 orelse good_eq u2 u1)
    | _ => false
  end

(* theory -> op2 -> term -> term -> bool *)
fun has_subterm_property thy oper bound_u body_u =
  exists (is_subterm_equation thy oper bound_u) (phrases oper body_u)

(* rep -> scope -> rep NameTable.table -> term -> term *)
fun choose_reps_in_term _ _ _ _ (Cst (Unity, T, _)) = Cst (Unity, T, Unit)
  | choose_reps_in_term _ _ _ _ (Cst (False, T, _)) = Cst (False, T, Formula)
  | choose_reps_in_term _ _ _ _ (Cst (True, T, _)) = Cst (True, T, Formula)
  | choose_reps_in_term _ _ {card_assigns, ...} _ (Cst (Num j, T, _)) =
    (case card_of_type card_assigns T of
       1 => if j = 0 then Cst (Unity, T, Unit)
            else Cst (Biggie, T, Opt (Atom 1))
     | k =>
       let val ok = (if T = int_T then atom_for_int k j <> ~1 else j < k) in
         if ok then Cst (Num j, T, Atom k) else Cst (Biggie, T, Opt (Atom k))
       end)
  | choose_reps_in_term _ _ scope _ (Cst (Biggie, T, _)) =
    Cst (Biggie, T, best_opt_set_rep_for_type scope T)
  | choose_reps_in_term _ _ {card_assigns, ...} _ (Cst (Suc, T, _)) =
    let val k = card_of_type card_assigns (domain_type T) in
      Cst (Suc, T, Func (if k = 1 then Unit else Atom k, Opt (Atom k)))
    end
  | choose_reps_in_term _ _ (scope as {card_assigns, ...}) _ (Cst (cst, T, _)) =
    if cst = Add orelse cst = Subtract orelse cst = Multiply then
      let
        val k = card_of_type card_assigns (domain_type T)
        val total = (cst = Subtract andalso T = nat_T)
      in
        Cst (cst, T, Func (Atom k, Func (Atom k, (not total ? Opt) (Atom k))))
      end
    else if cst = NatToInt then
      let
        val nat_card = card_of_type card_assigns @{typ nat}
        val int_card = card_of_type card_assigns @{typ int}
        val total = (max_int_for_card int_card >= nat_card + 1)
      in
        Cst (cst, T, Func (Atom nat_card, (not total ? Opt) (Atom int_card)))
      end
    else if cst = IntToNat then
      let
        val nat_card = card_of_type card_assigns @{typ nat}
        val int_card = card_of_type card_assigns @{typ int}
        val total = (max_int_for_card int_card < nat_card)
      in
        Cst (cst, T, Func (Atom int_card, (not total ? Opt) (Atom nat_card)))
      end
    else
      Cst (cst, T, best_non_opt_set_rep_for_type scope T)
  | choose_reps_in_term weak wish (scope as {card_assigns, ...}) table
                        (Op1 (oper, T, _, u1)) =
    let
      val wish1 = case oper of
                    Not => Formula
                  | _ => Any
      val u1 = choose_reps_in_term weak wish1 scope table u1
      val R1 = rep_of u1
      val R =
        (case oper of
           Not => formulate_rep R1
         | Finite => if is_opt_rep R1 then Opt (Atom 2) else Formula
         | Card => (if is_opt_rep R1
                       orelse exact_log 2 (card_of_rep R1)
                              >= card_of_type card_assigns nat_T then
                      best_opt_set_rep_for_type
                    else
                      best_non_opt_set_rep_for_type) scope T
         | First => (case R1 of
                       Struct [R11, _] => R11
                     | Opt (Struct [R11, _]) => opt_rep R11
                     | _ => raise Same)
         | Second => (case R1 of
                        Struct [_, R12] => R12
                      | Opt (Struct [_, R12]) => opt_rep R12
                      | _ => raise Same)
         | _ => raise Same)
        handle Same => (if is_opt_rep R1 then best_opt_set_rep_for_type
                        else best_non_opt_set_rep_for_type) scope T
    in optimize_unit (Op1 (oper, T, R, u1)) end
  | choose_reps_in_term weak wish (scope as {ctxt, card_assigns, datatypes})
                        table (Op2 (oper, T, _, u1, u2)) =
    if oper = All orelse oper = Exist then
      let
        val thy = ProofContext.theory_of ctxt
        val table' = fold (choose_rep_for_bound_var scope) (untuple I u1) table
        val u1' = choose_reps_in_term weak Any scope table' u1
        val u2' = choose_reps_in_term weak Formula scope table' u2
        val quant_u = Op2 (oper, T, formulate_rep (rep_of u2'), u1', u2')
      in
        if weak orelse is_precise_type datatypes (type_of u1)
           orelse has_subterm_property thy oper u1' u2' then
          quant_u
        else
          let
            val connective = if oper = All then And else Or
            val R1'' = opt_rep (rep_of u1')
            val table'' = NameTable.update (u1, R1'') table
            val u1'' = choose_reps_in_term weak Any scope table'' u1
            val u2'' = choose_reps_in_term weak Formula scope table'' u2
            val bottomed_u =
              substitute_in_term thy u1'' (Cst (Biggie, type_of u1'', R1''))
                                 u2''
          in
            Op2 (connective, T, min_rep (rep_of quant_u) (rep_of bottomed_u),
                 quant_u, bottomed_u)
          end
      end
    else if oper = Or orelse oper = Implies orelse oper = And then
      let
        val u1 = choose_reps_in_term weak Formula scope table u1
        val u2 = choose_reps_in_term weak Formula scope table u2
      in Op2 (oper, T, min_rep (rep_of u1) (rep_of u2), u1, u2) end
    else if oper = Less then
      let
        val u1 = choose_reps_in_term weak Any scope table u1
        val u2 = choose_reps_in_term weak Any scope table u2
        val R = if exists (is_opt_rep o rep_of) [u1, u2] then Opt (Atom 2)
                else Atom 2
      in Op2 (oper, T, R, u1, u2) end
    else if oper = StrictEq then
      Op2 (StrictEq, T, Formula, choose_reps_in_term weak Any scope table u1,
           choose_reps_in_term weak Any scope table u2)
    else if oper = Eq then
      let
        val thy = ProofContext.theory_of ctxt
        val wish' = best_set_rep_for_type scope (type_of u1)
        val u1 = choose_reps_in_term weak wish' scope table u1
        val u2 = choose_reps_in_term weak wish' scope table u2
        (* term -> op2 * rep *)
        fun odd_case u =
          if is_constructive thy u then (StrictEq, Formula)
          else (Eq, Opt (Atom 2))
        val (oper, R) =
          case (is_opt_rep (rep_of u1), is_opt_rep (rep_of u2)) of
            (true, true) => (Eq, Opt (Atom 2))
          | (true, false) => odd_case u1
          | (false, true) => odd_case u2
          | (false, false) => (Eq, Formula)
      in Op2 (oper, T, R, u1, u2) end
    else if oper = Subset then
      let
        val wish' = best_set_rep_for_type scope (type_of u1)
        val u1 = choose_reps_in_term weak wish' scope table u1
        val u2 = choose_reps_in_term weak wish' scope table u2
        val R = if exists (is_opt_rep o rep_of) [u1, u2] then Opt (Atom 2)
                else Formula
      in Op2 (oper, T, R, u1, u2) end
    else if oper = Apply then
      let
        val thy = ProofContext.theory_of ctxt
        val u1 = choose_reps_in_term weak
                     (best_set_rep_for_type scope (type_of u1)) scope table u1
        val R1 = rep_of u1
        val ran_R = range_rep (fn () => card_of_type card_assigns T) R1
        val u2 = choose_reps_in_term weak (best_domain_rep R1) scope table u2
        val R = ((ran_R <> Unit andalso is_opt_rep (rep_of u2)) ? opt_rep) ran_R
      in optimize_unit (optimize_biggie thy (Op2 (Apply, T, R, u1, u2))) end
    else if oper = Abstraction then
      let
        val R = best_set_rep_for_type scope T
        val R1 = domain_rep R
        val R2 = range_rep (K 0) R
        val table = NameTable.update (u1, R1) table
        val u1 = choose_reps_in_term weak R1 scope table u1
        val u2 = choose_reps_in_term weak R2 scope table u2
        val R = (if is_opt_rep (rep_of u2) then opt_rep else unopt_rep) R
      in optimize_unit (Op2 (Abstraction, T, R, u1, u2)) end
    else
      let
        val u1 = choose_reps_in_term weak wish scope table u1
        val R1 = rep_of u1
        val u2 = choose_reps_in_term weak R1 scope table u2
        val R2 = rep_of u2
        val R = if R1 = R2 then R1 else best_set_rep_for_type scope T
      in optimize_unit (Op2 (oper, T, R, u1, u2)) end
  | choose_reps_in_term weak wish (scope as {ctxt, ...}) table
                        (Op3 (Let, T, _, u1, u2, u3)) =
    let
      val thy = ProofContext.theory_of ctxt
      val u2 = choose_reps_in_term weak Any scope table u2
      val R2 = rep_of u2
      val u1 = change_name_rep u1 R2
      val table = NameTable.update (u1, R2) table
      val u3 = choose_reps_in_term weak wish scope table u3
    in optimize_unit (optimized_let thy u1 u2 u3) end
  | choose_reps_in_term weak wish scope table (Op3 (If, T, _, u1, u2, u3)) =
    let
      val u1 = choose_reps_in_term weak Formula scope table u1
      val u2 = choose_reps_in_term weak wish scope table u2
      val u3 = choose_reps_in_term weak (rep_of u2) scope table u3
      val min_R = min_rep (rep_of u2) (rep_of u3)
    in
      optimize_unit (Op3 (If, T, (is_opt_rep (rep_of u1) ? opt_rep) min_R,
                          u1, u2, u3))
    end
  | choose_reps_in_term weak wish scope table (Tuple (T, _, us)) =
    let
      val Rs = map (best_one_rep_for_type scope o type_of) us
      val us = map2 (fn R => choose_reps_in_term weak R scope table) Rs us
      val R = if forall (equal Unit) Rs then Unit else Struct Rs
      val R' = if exists (is_opt_rep o rep_of) us then opt_rep R else R
    in optimize_unit (Tuple (T, R', us)) end
  | choose_reps_in_term _ _ _ table v =
    optimize_unit (change_name_rep v (lookup_name table v))

(* Kodkod output *)

(* int -> Kodkod.n_ary_index list -> Kodkod.n_ary_index list
   -> int * Kodkod.n_ary_index list *)
fun fresh_n_ary_index n [] ys = (0, (n, 1) :: ys)
  | fresh_n_ary_index n ((m, j) :: xs) ys =
    if m = n then (j, ys @ ((m, j + 1) :: xs))
    else fresh_n_ary_index n xs ((m, j) :: ys)
(* int -> name_pool -> int * name_pool *)
fun fresh_rel n {rels, vars, formula_reg, rel_reg} =
  let val (j, rels') = fresh_n_ary_index n rels [] in
    (j, {rels = rels', vars = vars, formula_reg = formula_reg,
         rel_reg = rel_reg})
  end
(* int -> name_pool -> int * name_pool *)
fun fresh_var n {rels, vars, formula_reg, rel_reg} =
  let val (j, vars') = fresh_n_ary_index n vars [] in
    (j, {rels = rels, vars = vars', formula_reg = formula_reg,
         rel_reg = rel_reg})
  end
(* int -> name_pool -> int * name_pool *)
fun fresh_formula_reg {rels, vars, formula_reg, rel_reg} =
  (formula_reg, {rels = rels, vars = vars, formula_reg = formula_reg + 1,
                 rel_reg = rel_reg})
(* int -> name_pool -> int * name_pool *)
fun fresh_rel_reg {rels, vars, formula_reg, rel_reg} =
  (rel_reg, {rels = rels, vars = vars, formula_reg = formula_reg,
             rel_reg = rel_reg + 1})

(* term -> term list * name_pool * term NameTable.table
   -> term list * name_pool * term NameTable.table *)
fun rename_plain_var v (ws, pool, table) =
  let
    val is_formula = (rep_of v = Formula)
    val fresh = if is_formula then fresh_formula_reg else fresh_rel_reg
    val (j, pool) = fresh pool
    val ctor = if is_formula then FormulaReg else RelReg
    val w = ctor (j, type_of v, rep_of v)
  in (w :: ws, pool, NameTable.update (v, w) table) end

(* typ -> rep -> term list -> term *)
fun shape_tuple (T as Type ("*", [T1, T2])) (R as Struct [R1, R2]) us =
    let val arity1 = arity_of_rep R1 in
      Tuple (T, R, [shape_tuple T1 R1 (List.take (us, arity1)),
                    shape_tuple T2 R2 (List.drop (us, arity1))])
    end
  | shape_tuple T R [u] = if type_of u = T then u else Tuple (T, R, [u])
  | shape_tuple T R us = Tuple (T, R, us)

(* bool -> term -> term list * name_pool * term NameTable.table
   -> term list * name_pool * term NameTable.table *)
fun rename_n_ary_var rename_free v (ws, pool, table) =
  let
    val T = type_of v
    val R = rep_of v
    val arity = arity_of_rep R
    val nick = nickname_of v
    val (ctor, fresh) = if rename_free then (FreeRel, fresh_rel)
                                       else (BoundRel, fresh_var)
  in
    if not (rename_free) andalso arity > 1 then
      let
        val atom_schema = atom_schema_of_rep R
        val type_schema = type_schema_of_rep T R
        val (js, pool) = funpow arity (fn (js, pool) =>
                                          let val (j, pool) = fresh 1 pool in
                                            (j :: js, pool)
                                          end)
                                ([], pool)
        val ws' = map3 (fn j => fn k => fn T =>
                           ctor ((1, j), T, Atom k,
                                 nick ^ " [" ^ Int.toString j ^ "]"))
                       (rev js) atom_schema type_schema
      in (ws' @ ws, pool, NameTable.update (v, shape_tuple T R ws') table) end
    else
      let
        val (j, pool) = fresh arity pool
        val w = ctor ((arity, j), T, R, nick)
      in (w :: ws, pool, NameTable.update (v, w) table) end
  end

(* term list -> name_pool -> term NameTable.table
  -> term list * name_pool * term NameTable.table *)
fun rename_free_vars vs pool table =
  let
    val vs = filter (not_equal Unit o rep_of) vs
    val (vs, pool, table) = fold (rename_n_ary_var true) vs ([], pool, table)
  in (rev vs, pool, table) end

(* name_pool -> term NameTable.table -> term -> term *)
fun rename_vars_in_term pool table u =
  case u of
    Cst _ => u
  | Op1 (oper, T, R, u1) => Op1 (oper, T, R, rename_vars_in_term pool table u1)
  | Op2 (oper, T, R, u1, u2) =>
    if oper = All orelse oper = Exist orelse oper = Abstraction then
      let
        val (_, pool, table) = fold (rename_n_ary_var false) (untuple I u1)
                                    ([], pool, table)
      in
        Op2 (oper, T, R, rename_vars_in_term pool table u1,
             rename_vars_in_term pool table u2)
      end
    else
      Op2 (oper, T, R, rename_vars_in_term pool table u1,
           rename_vars_in_term pool table u2)
  | Op3 (Let, T, R, u1, u2, u3) =>
    if rep_of u2 = Unit orelse inline_term u2 then
      let
        val u2 = rename_vars_in_term pool table u2
        val table = NameTable.update (u1, u2) table
      in rename_vars_in_term pool table u3 end
    else
      let
        val bs = untuple I u1
        val (_, pool, table') = fold rename_plain_var bs ([], pool, table)
        val u11 = rename_vars_in_term pool table' u1
      in
        Op3 (Let, T, R, rename_vars_in_term pool table' u1,
             rename_vars_in_term pool table u2,
             rename_vars_in_term pool table' u3)
      end
  | Op3 (oper, T, R, u1, u2, u3) =>
    Op3 (oper, T, R, rename_vars_in_term pool table u1,
         rename_vars_in_term pool table u2, rename_vars_in_term pool table u3)
  | Tuple (T, R, us) => Tuple (T, R, map (rename_vars_in_term pool table) us)
  | _ => lookup_name table u

end;
